# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Models used by debusine client."""

import datetime as dt
import re
from collections.abc import Collection, Iterator, Sequence
from enum import StrEnum
from pathlib import Path
from typing import Annotated, Any, Literal, NewType, Self, TypeVar, override

import pydantic
import yaml

from debusine.artifacts.models import (
    CollectionCategory,
    DebusineTaskConfiguration,
)
from debusine.assets import AssetCategory, asset_data_model
from debusine.utils import calculate_hash


class StrictBaseModel(pydantic.BaseModel):
    """Stricter pydantic configuration."""

    model_config = pydantic.ConfigDict(validate_assignment=True)


class Link(StrictBaseModel):
    """
    Structured version of a resource link.

    This is modeled as a subset of
    https://developer.mozilla.org/en-US/docs/Web/HTML/Reference/Elements/link
    """  # noqa: E501

    #: Resource URL
    href: str
    type: str | None = None

    def can_be_str(self) -> bool:
        """Check if the only field set is href."""
        return self.type is None


class ServerObject(StrictBaseModel):
    """Model representing an object on the server."""

    # See https://en.wikipedia.org/wiki/HATEOAS

    #: URLs to related resources
    links: dict[str, str | Link] = {}

    def get_link_url(self, name: str) -> str:
        """Return the href URL for a link."""
        res = self.links[name]
        if isinstance(res, str):
            return res
        return res.href


class PaginatedResponse(StrictBaseModel):
    """Paginated response from the API."""

    count: int | None = None
    next: pydantic.AnyUrl | None = None
    previous: pydantic.AnyUrl | None = None
    results: list[dict[str, Any]]


class WorkRequestRequest(StrictBaseModel):
    """Client send a WorkRequest to the server."""

    task_name: str
    workspace: str | None = None
    task_data: dict[str, Any]
    event_reactions: dict[str, Any]


class WorkRequestResponse(ServerObject):
    """Server return a WorkRequest to the client."""

    id: int
    url: str
    parent: int | None = None  # Server > 0.14.7 specifies parent
    workflow_root: int | None = None  # Server > 0.14.7 specifies workflow_root
    created_at: dt.datetime
    started_at: dt.datetime | None = None
    completed_at: dt.datetime | None = None
    duration: float | None = None
    status: str
    result: str
    worker: int | None = None
    task_type: str
    task_name: str
    task_data: dict[str, Any]
    dynamic_task_data: dict[str, Any] | None = None
    priority_base: int
    priority_adjustment: int
    artifacts: list[int]
    scope: str | None = None  # Server >= 0.12 specifies scope
    workspace: str
    scheduler_tags_provided: list[str] = []
    scheduler_tags_required: list[str] = []
    token: str | None = None  # Worker receives a work request token

    @override
    def __str__(self) -> str:
        return f'WorkRequest: {self.id}'


class WorkRequestExternalDebsignRequest(StrictBaseModel):
    """Client sends data from an external `debsign` run to the server."""

    signed_artifact: int


class OnWorkRequestCompleted(ServerObject):
    """
    Server return an OnWorkRequestCompleted to the client.

    Returned via websocket consumer endpoint.
    """

    work_request_id: int
    completed_at: dt.datetime
    result: str


class RuntimeParameter(StrEnum):
    """WorkflowTemplate ``runtime_parameter`` special values."""

    ANY = "any"


RuntimeParameters = (
    dict[str, list[str] | Literal[RuntimeParameter.ANY]]
    | Literal[RuntimeParameter.ANY]
)


class WorkflowTemplateDataNew(StrictBaseModel):
    """Data for a WorkflowTemplate to be created."""

    name: str
    task_name: str
    static_parameters: dict[str, Any]
    runtime_parameters: RuntimeParameters = {}
    priority: int = 0
    extra_groups: list[int | str] = []
    restricted: bool = False


class WorkflowTemplateData(ServerObject):
    """Data for a WorkflowTemplate."""

    id: int
    name: str
    task_name: str
    static_parameters: dict[str, Any]
    runtime_parameters: RuntimeParameters
    priority: int = 0
    extra_groups: list[int | str] = []
    restricted: bool = False


class CreateWorkflowRequest(StrictBaseModel):
    """Client sends a workflow creation request to the server."""

    template_name: str = ""
    task_name: str = ""
    workspace: str | None = None
    task_data: dict[str, Any]

    @pydantic.model_validator(mode="after")
    def validate_names(self) -> Self:
        """Exactly one of ``template_name`` and ``task_name`` must be set."""
        if bool(self.template_name) == bool(self.task_name):
            raise ValueError(
                "Exactly one of template_name and task_name must be set"
            )
        return self


class FileRequest(StrictBaseModel):
    """Declare a FileRequest: client sends it to the server."""

    size: pydantic.NonNegativeInt
    checksums: dict[str, Annotated[str, pydantic.Field(max_length=255)]]
    type: Literal["file"]
    #: A media type, suitable for an HTTP ``Content-Type`` header.
    content_type: str | None = None

    @staticmethod
    def create_from(
        path: Path, *, content_type: str | None = None
    ) -> "FileRequest":
        """Return a FileRequest for the file path."""
        return FileRequest(
            size=path.stat().st_size,
            checksums={"sha256": calculate_hash(path, "sha256").hex()},
            type="file",
            content_type=content_type,
        )


class FileResponse(ServerObject):
    """Declare a FileResponse: server sends it to the client."""

    size: pydantic.NonNegativeInt
    checksums: dict[str, Annotated[str, pydantic.Field(max_length=255)]]
    type: Literal["file"]
    url: pydantic.AnyUrl
    content_type: str | None = None


FilesRequestType = NewType("FilesRequestType", dict[str, FileRequest])
FilesResponseType = NewType("FilesResponseType", dict[str, FileResponse])


class ArtifactCreateRequest(StrictBaseModel):
    """Declare an ArtifactCreateRequest: client sends it to the server."""

    category: str
    workspace: str | None = None
    files: FilesRequestType = FilesRequestType({})
    data: dict[str, Any] = {}
    work_request: int | None = None
    expire_at: dt.datetime | None = None


class ArtifactResponse(ServerObject):
    """Declare an ArtifactResponse: server sends it to the client."""

    id: int
    url: str
    scope: str
    workspace: str
    category: str
    created_at: dt.datetime
    data: dict[str, Any]
    download_tar_gz_url: pydantic.AnyUrl
    files_to_upload: list[str]
    expire_at: dt.datetime | None = None
    files: FilesResponseType = FilesResponseType({})


class RemoteArtifact(ServerObject):
    """Declare RemoteArtifact."""

    id: int
    url: str
    workspace: str


class AssetCreateRequest(StrictBaseModel):
    """Request for the Asset creation API."""

    category: AssetCategory
    data: dict[str, Any]
    work_request: int | None = None
    workspace: str


class AssetResponse(ServerObject):
    """Response from an Asset creation / listing API."""

    id: int
    category: AssetCategory
    data: dict[str, Any]
    work_request: int | None = None
    workspace: str

    @pydantic.model_validator(mode="after")
    def validate_data(self) -> Self:
        """Validate data using the correct data model."""
        asset_data_model(self.category, self.data)
        return self


class AssetsResponse(
    pydantic.RootModel[Sequence[AssetResponse]], StrictBaseModel
):
    """A response from the server with multiple AssetResponse objects."""

    model_config = pydantic.ConfigDict(validate_assignment=True)

    @override
    def __iter__(self) -> Iterator[AssetResponse]:  # type: ignore[override]
        """Iterate over individual asset responses."""
        return iter(self.root)


class AssetPermissionCheckResponse(ServerObject):
    """A response from the asset permission check endpoint."""

    has_permission: bool
    username: str | None
    user_id: int | None
    resource: dict[str, Any] | None


class RelationType(StrEnum):
    """Possible values for `RelationCreateRequest.type`."""

    EXTENDS = "extends"
    RELATES_TO = "relates-to"
    BUILT_USING = "built-using"


class RelationCreateRequest(StrictBaseModel):
    """Declare a RelationCreateRequest: client sends it to the server."""

    artifact: int
    target: int
    type: RelationType


class RelationResponse(ServerObject):
    """Declare a RelationResponse."""

    id: int
    artifact: int
    target: int
    type: RelationType


class RelationsResponse(
    pydantic.RootModel[Sequence[RelationResponse]], StrictBaseModel
):
    """A response from the server with multiple RelationResponse objects."""

    @override
    def __iter__(self) -> Iterator[RelationResponse]:  # type: ignore[override]
        """Iterate over individual relation responses."""
        return iter(self.root)


class CollectionRelationType(StrEnum):
    """Possible values for `CollectionRelationResponse.type`."""

    SUITE_FORKED_FROM = "suite-forked-from"
    SUITE_BASED_ON = "suite-based-on"
    SUITE_REQUIRES = "suite-requires"
    SUITE_TARGETING = "suite-targeting"
    SUITE_DEFAULT_QA_RESULTS = "suite-default-qa-results"


class CollectionIdentifierRequest(StrictBaseModel):
    """Identifies a collection in API requests."""

    name: str
    category: CollectionCategory


class CollectionRelationSetRequest(StrictBaseModel):
    """Used to set all targets for a collection relation type."""

    targets: list[CollectionIdentifierRequest]


class CollectionIdentifierResponse(StrictBaseModel):
    """Identifies a collection in API responses."""

    name: str
    category: CollectionCategory


class CollectionRelationSetResponse(StrictBaseModel):
    """Updated targets for a collection relation list."""

    targets: list[CollectionIdentifierResponse]


class CollectionRelationResponse(StrictBaseModel):
    """Relationship between two collections."""

    source_collection: CollectionIdentifierResponse
    target_collection: CollectionIdentifierResponse
    type: CollectionRelationType
    position: int | None = None


class LookupChildType(StrEnum):
    """Possible values for `LookupDict.child_type` and `expect_type`."""

    BARE = "bare"
    ARTIFACT = "artifact"
    ARTIFACT_OR_PROMISE = "artifact-or-promise"
    COLLECTION = "collection"
    ANY = "any"


class LookupResultType(StrEnum):
    """A collection item type returned by a lookup."""

    BARE = "bare"
    ARTIFACT = "artifact"
    COLLECTION = "collection"


LookupMultiple = list[int | str | dict[str, Any]]
LookupSingle = int | str


class LookupSingleRequest(StrictBaseModel):
    """A request from the client to look up a single collection item."""

    lookup: LookupSingle
    work_request: int | None = None
    expect_type: LookupChildType = LookupChildType.ANY
    default_category: str | None = None
    workspace: str


class LookupMultipleRequest(StrictBaseModel):
    """A request from the client to look up multiple collection items."""

    lookup: LookupMultiple
    work_request: int | None = None
    expect_type: LookupChildType
    default_category: str | None = None
    workspace: str


class CollectionItemResponse(StrictBaseModel):
    """A collection item in a lookup response."""

    name: str
    data: dict[str, Any]


class LookupSingleResponse(ServerObject):
    """A response from the server with a single lookup result."""

    result_type: LookupResultType
    collection_item: CollectionItemResponse | None = None
    artifact: int | None = None
    collection: int | None = None
    parent_collection_lookup: LookupSingle | None = None

    @override
    def __eq__(self, other: Any) -> bool:
        """Treat our subclasses equally when evaluating equality."""
        if not isinstance(other, LookupSingleResponse):
            return False
        return self.model_dump() == other.model_dump()


class LookupSingleResponseArtifact(LookupSingleResponse):
    """
    A response from the server with a single lookup result for an artifact.

    Used to assist type annotations.
    """

    result_type: Literal[LookupResultType.ARTIFACT] = LookupResultType.ARTIFACT
    artifact: int


class LookupSingleResponseCollection(LookupSingleResponse):
    """
    A response from the server with a single lookup result for a collection.

    Used to assist type annotations.
    """

    result_type: Literal[LookupResultType.COLLECTION] = (
        LookupResultType.COLLECTION
    )
    collection: int


# TODO: We should use PEP 695 type parameter declarations here instead, but
# that causes "Overloaded function implementation cannot produce return
# type" errors on Debusine.lookup_multiple.
LSR = TypeVar("LSR", bound=LookupSingleResponse, covariant=True)


class LookupMultipleResponse(
    pydantic.RootModel[Sequence[LSR]], StrictBaseModel
):
    """A response from the server with multiple lookup results."""

    @override
    def __iter__(self) -> Iterator[LSR]:  # type: ignore[override]
        """Iterate over individual results."""
        return iter(self.root)


re_nonce = re.compile(r"^[A-Za-z0-9_-]{8,64}$")
re_challenge = re.compile(r"^\w{4,10}(?: \w{4,10}){2,7}$", re.ASCII)


class EnrollPayload(pydantic.BaseModel):
    """Client-provided enrollment payload."""

    model_config = pydantic.ConfigDict(validate_assignment=True, extra="forbid")

    #: Nonce identifying the client enrollment
    nonce: Annotated[str, pydantic.Field(pattern=re_nonce)]
    #: Human-readable challenge to verify the client in the web UI
    challenge: Annotated[str, pydantic.Field(pattern=re_challenge)]
    #: Scope. Informational, the confirmation page will warn if invalid
    scope: str
    #: Hostname. Informational, shown in the confirmation page and in the
    #: generated token comment
    hostname: str


re_token = re.compile(r"^[A-Za-z0-9]{8,64}$")


class EnrollOutcome(StrEnum):
    """User action in response to an enroll confirmation request."""

    CONFIRM = "confirm"
    CANCEL = "cancel"


class EnrollConfirmPayload(pydantic.BaseModel):
    """Enrollment response from the server."""

    model_config = pydantic.ConfigDict(validate_assignment=True, extra="forbid")

    outcome: EnrollOutcome
    token: Annotated[str, pydantic.Field(pattern=re_token)] | None = None


class TaskConfigurationCollection(ServerObject):
    """Response fragment identifying a TaskConfiguration collection."""

    id: int
    name: str
    data: dict[str, Any]


class TaskConfigurationCollectionContents(ServerObject):
    """Bundle together a TaskConfiguration collection and its items."""

    collection: TaskConfigurationCollection
    items: list[DebusineTaskConfiguration]


class TaskConfigurationCollectionUpdateResults(ServerObject):
    """Results of a task configuration collection update."""

    added: int
    updated: int
    removed: int
    unchanged: int


class WorkspaceInheritanceChainElement(StrictBaseModel):
    """An element in an inheritance chain."""

    id: int | None = None
    scope: str | None = None
    workspace: str | None = None

    @pydantic.model_validator(mode="after")
    def not_empty(self) -> Self:
        """Ensure at least one of id or workspace is set."""
        if self.id is None and self.workspace is None:
            raise ValueError("at least id or workspace need to be set")
        return self

    def matches(self, element: "WorkspaceInheritanceChainElement") -> bool:
        """Check if the components of element that are set match this one."""
        return (
            (element.id is None or element.id == self.id)
            and (element.scope is None or element.scope == self.scope)
            and (
                element.workspace is None or element.workspace == self.workspace
            )
        )

    @classmethod
    def from_string(cls, value: str) -> Self:
        """Resolve a command line argument to a WorkspaceInheritanceChain."""
        if value.isdigit():
            return cls(id=int(value))
        elif "/" in value:
            scope, workspace = value.split("/", 1)
            return cls(scope=scope, workspace=workspace)
        else:
            return cls(workspace=value)


class WorkspaceInheritanceChain(ServerObject):
    """Ordered list of workspaces inherited by a workspace."""

    chain: list[WorkspaceInheritanceChainElement] = []

    def __add__(self, other: "WorkspaceInheritanceChain") -> Self:
        """Return the concatenation of this chain and another."""
        return self.__class__(chain=self.chain + other.chain)

    def __sub__(self, other: "WorkspaceInheritanceChain") -> Self:
        """Return a copy of this chain without the elements in other."""
        filtered: list[WorkspaceInheritanceChainElement] = []
        for element in self.chain:
            for to_remove in other.chain:
                if element.matches(to_remove):
                    break
            else:
                filtered.append(element)

        return self.__class__(chain=filtered)

    @classmethod
    def from_strings(cls, values: Collection[str]) -> Self:
        """Build an inheritance chain from a list of user-provided strings."""
        return cls(
            chain=[
                WorkspaceInheritanceChainElement.from_string(value)
                for value in values
            ]
        )


class CollectionDataNew(StrictBaseModel):
    """Data for a Collection to be created."""

    name: str
    category: str
    full_history_retention_period: int | None = None
    metadata_only_retention_period: int | None = None
    data: dict[str, Any]


class CollectionData(ServerObject):
    """Data for a Collection."""

    id: int
    name: str
    category: str
    full_history_retention_period: int | None = None
    metadata_only_retention_period: int | None = None
    data: dict[str, Any]
    stats: list[dict[str, Any]] | None = None


class CollectionRoles(StrEnum):
    """Available roles for a Collection."""

    # The database model also defines OWNER, but we don't have a use case
    # for clients granting that yet.  Omit it for now to avoid confusing
    # people.
    VIEWER = "viewer"


class CollectionTokenGrantRequest(ServerObject):
    """Request to grant a Collection role to a token or an external username."""

    token_hash: str | None = None
    username: str | None = None
    comment: str = ""
    role: CollectionRoles

    @pydantic.model_validator(mode="after")
    def validate_token_or_username(self) -> Self:
        """At least one of ``token_hash`` and ``username`` must be set."""
        if self.token_hash is None and self.username is None:
            raise ValueError(
                "At least one of token_hash and username must be set"
            )
        return self


class CollectionTokenGrantResponse(ServerObject):
    """Response for a token granted a Collection role."""

    token_type: str
    token_hash: str
    token_key: str | None = None
    username: str
    comment: str
    role: CollectionRoles


class EmitMetricRequest(StrictBaseModel):
    """A request for the server to emit a metric on behalf of the client."""

    metric_type: Literal["counter", "gauge", "summary", "histogram"]
    name: str
    labels: dict[str, str]
    value: float


# Workarounds for https://github.com/yaml/pyyaml/issues/722
yaml.SafeDumper.add_multi_representer(
    StrEnum,
    yaml.representer.SafeRepresenter.represent_str,
)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the package upload workflow."""

from collections.abc import Sequence
from typing import Any, ClassVar, override
from unittest.mock import call, patch

from debusine.artifacts.models import (
    ArtifactCategory,
    BareDataCategory,
    CollectionCategory,
    DebianUpload,
    DebusinePromise,
    TaskTypes,
)
from debusine.client.models import LookupChildType, RuntimeParameter
from debusine.db.models import (
    Artifact,
    Collection,
    CollectionItem,
    WorkRequest,
    WorkflowTemplate,
    default_workspace,
)
from debusine.server.workflows import (
    PackageUploadWorkflow,
    WorkflowValidationError,
)
from debusine.server.workflows.base import orchestrate_workflow
from debusine.server.workflows.models import (
    PackageUploadWorkflowDynamicData,
    WorkRequestWorkflowData,
)
from debusine.server.workflows.tests.helpers import WorkflowTestCase
from debusine.tasks.models import (
    ActionUpdateCollectionWithArtifacts,
    InputArtifactMultiple,
    InputArtifactSingle,
    LookupMultiple,
    SbuildData,
    SbuildInput,
)


class PackageUploadWorkflowTests(WorkflowTestCase[PackageUploadWorkflow]):
    """Unit tests for :py:class:`PackageUploadWorkflow`."""

    SAMPLE_TASK_DATA = {
        "target": "sftp://upload.example.org/queue/",
        "vendor": "debian",
        "codename": "trixie",
        "require_signature": False,
    }

    source_artifact: ClassVar[Artifact]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.source_artifact = (
            cls.playground.create_minimal_source_package_artifact()
        )

    @override
    def setUp(self) -> None:
        super().setUp()
        self.task_data["source_artifact"] = self.source_artifact.id

    def create_source_binary_artifacts(
        self, architectures: Sequence[str] | None = None
    ) -> tuple[Artifact, list[Artifact], Collection]:
        """Return one source and a list of binary packages (one per arch)."""
        if architectures is None:
            architectures = ("amd64",)
        collection = self.playground.create_collection(
            "test", CollectionCategory.TEST
        )

        source = self.playground.create_minimal_source_package_artifact()
        uploads = []
        for architecture in architectures:
            upload = self.playground.create_upload_artifacts(
                source=False,
                binary=True,
                binaries=[("hello", architecture)],
            )
            uploads.append(upload.upload)
            for binary in upload.binaries:
                CollectionItem.objects.create_from_artifact(
                    binary,
                    parent_collection=collection,
                    name=f"binary-{binary.data['deb_fields']['Architecture']}",
                    created_by_user=self.playground.get_default_user(),
                    data={},
                    created_by_workflow=None,
                )

        return source, uploads, collection

    def test_validate_input_source_binary_artifacts_not_set(self) -> None:
        """Invalid: source_artifact and binaries_artifacts not set."""
        del self.task_data["source_artifact"]
        self.validate = False
        with self.assertRaisesRegex(
            WorkflowValidationError,
            '"source_artifact" or "binary_artifacts" must be set',
        ):
            self.task.validate_input()

    def test_validate_input_source_package_vendor_codename_not_set(
        self,
    ) -> None:
        """Raise WorkflowValidationError: vendor and codename required."""
        self.validate = False
        del self.task_data["vendor"]
        del self.task_data["codename"]

        with self.assertRaisesRegex(
            WorkflowValidationError,
            '"vendor" and "codename" are required when '
            'source artifact category is debian:source-package',
        ):
            self.task.validate_input()

    def test_create_orchestrator(self) -> None:
        """A PackageUploadWorkflow can be instantiated."""
        source_artifact, [binary_artifact], _ = (
            self.create_source_binary_artifacts(["arm64"])
        )
        target_distribution = "debian:bookworm"
        self.task_data.update(
            {
                "source_artifact": source_artifact.id,
                "binary_artifacts": [binary_artifact.id],
                "target_distribution": target_distribution,
            }
        )
        w = self.task
        self.assertEqual(w.data.source_artifact, source_artifact.id)
        self.assertEqual(
            w.data.binary_artifacts, LookupMultiple((binary_artifact.id,))
        )

        self.assertEqual(w.data.target_distribution, target_distribution)

    def orchestrate(self, *, data: dict[str, Any]) -> WorkRequest:
        """Create a PackageUpload workflow and call orchestrate_workflow."""
        template = WorkflowTemplate.objects.create(
            name="package_upload",
            workspace=default_workspace(),
            task_name="package_upload",
            runtime_parameters=RuntimeParameter.ANY,
        )

        wr = self.playground.create_workflow(
            template,
            task_data={"target": "sftp://upload.example.org/queue", **data},
        )
        self.assertTrue(orchestrate_workflow(wr))

        return wr

    def test_populate_makesourcepackageupload(self) -> None:
        """
        Workflow create two children work request.

        Work requests created:
        -makesourcepackageupload
        -package_upload
        """
        source = self.playground.create_minimal_source_package_artifact()

        with patch.object(
            PackageUploadWorkflow,
            "requires_artifact",
            wraps=PackageUploadWorkflow.requires_artifact,
        ) as requires_artifact:
            wr = self.orchestrate(
                data={
                    "source_artifact": source.id,
                    "target_distribution": "debian:bookworm",
                    "since_version": "1.0",
                    "require_signature": False,
                    "vendor": "debian",
                    "codename": "bullseye",
                }
            )

        self.assertEqual(wr.children.count(), 2)

        [make_source] = wr.children.filter(task_name="makesourcepackageupload")

        self.assertEqual(
            make_source.workflow_data_json,
            {
                "display_name": "Make .changes file",
                "step": "source-package-upload",
            },
        )

        self.assertEqual(
            make_source.task_data,
            {
                "build_architecture": "amd64",
                "environment": "debian/match:codename=bullseye",
                "input": {"source_artifact": source.id},
                "since_version": "1.0",
                "target_distribution": "debian:bookworm",
            },
        )

        self.assertEqual(make_source.parent, wr)

        self.assert_work_request_event_reactions(
            make_source,
            on_success=[
                ActionUpdateCollectionWithArtifacts(
                    artifact_filters={"category": "debian:upload"},
                    collection="internal@collections",
                    name_template="package-upload-source",
                )
            ],
        )

        # Assert that self.requires_artifact() was called.
        # It cannot test the dependencies because source_artifact was
        # a real artifact and not a promise (so no dependencies were added)
        self.assertIn(
            call(make_source, source.id), requires_artifact.mock_calls
        )

        [package_upload] = wr.children.filter(task_name="packageupload")
        self.assertEqual(package_upload.task_type, TaskTypes.SERVER)
        self.assertEqual(
            package_upload.workflow_data_json,
            {
                "display_name": (
                    "Package upload internal@collections/"
                    "name:package-upload-source"
                ),
                "step": (
                    "package-upload-internal@collections/"
                    "name:package-upload-source"
                ),
            },
        )
        self.assertEqual(
            package_upload.task_data,
            {
                "input": {
                    "upload": "internal@collections/name:package-upload-source"
                },
                "target": "sftp://upload.example.org/queue",
                "delayed_days": None,
            },
        )

        self.assertQuerySetEqual(
            package_upload.dependencies.all(),
            list(
                WorkRequest.objects.filter(task_name="makesourcepackageupload")
            ),
        )

    def test_populate_source_is_package_upload(self) -> None:
        """Category source package is UPLOAD: skip makesourcepackageupload."""
        source = self.playground.create_upload_artifacts().upload

        wr = self.orchestrate(
            data={
                "source_artifact": source.id,
                "target_distribution": "debian:bookworm",
                "since_version": "1.0",
                "require_signature": False,
            }
        )

        self.assertEqual(wr.children.count(), 1)

        [package_upload] = wr.children.filter(task_name="packageupload")
        self.assertEqual(
            package_upload.workflow_data_json,
            {
                "display_name": f"Package upload {source.id}",
                "step": f"package-upload-{source.id}",
            },
        )
        self.assertEqual(
            package_upload.task_data,
            {
                "input": {"upload": source.id},
                "target": "sftp://upload.example.org/queue",
                "delayed_days": None,
            },
        )

    def test_populate_package_uploads_binaries_only(self) -> None:
        """The workflow create children package uploads: binaries only."""
        _, [binary], _ = self.create_source_binary_artifacts()

        wr = self.orchestrate(
            data={
                "binary_artifacts": [binary.id],
                "target_distribution": "debian:bookworm",
                "require_signature": False,
            },
        )

        self.assertEqual(
            wr.children.filter(task_name="packageupload").count(), 1
        )

        # Other properties of the created package is tested on different
        # unit tests

    def test_populate_merge_uploads(self) -> None:
        """Test mergeuploads task is created (merge_uploads=True)."""
        parent = self.playground.create_workflow()
        assert parent.internal_collection is not None

        architectures = ("amd64", "i386")
        source = self.playground.create_minimal_source_package_artifact()
        for architecture in architectures:
            sbuild = parent.create_child_worker(
                task_name="sbuild",
                task_data=SbuildData(
                    input=SbuildInput(source_artifact=source.id),
                    build_architecture=architecture,
                    environment="debian/match:codename=bookworm",
                ),
                workflow_data=WorkRequestWorkflowData(
                    display_name=architecture, step=architecture
                ),
            )
            self.playground.create_bare_data_item(
                parent.internal_collection,
                f"build-{architecture}",
                category=BareDataCategory.PROMISE,
                data=DebusinePromise(
                    promise_work_request_id=sbuild.id,
                    promise_workflow_id=parent.id,
                    promise_category=ArtifactCategory.UPLOAD,
                ),
            )

        wr = self.playground.create_workflow(
            task_name="package_upload",
            task_data={
                "source_artifact": source.id,
                "binary_artifacts": {
                    "collection": "internal@collections",
                    "child_type": LookupChildType.ARTIFACT_OR_PROMISE,
                    "name__startswith": "build-",
                },
                "require_signature": False,
                "target": "sftp://upload.example.org/queue",
                "target_distribution": "debian/match:codename=bookworm",
                "merge_uploads": True,
                "vendor": "debian",
                "codename": "bullseye",
            },
            parent=parent,
        )
        self.assertTrue(orchestrate_workflow(wr))

        self.assertEqual(
            wr.children.filter(task_name="mergeuploads").count(), 1
        )

        [merge_uploads] = wr.children.filter(task_name="mergeuploads")

        self.assertEqual(
            merge_uploads.task_data,
            {
                "input": {
                    "uploads": [
                        "internal@collections/name:package-upload-source"
                    ]
                    + [
                        f"internal@collections/name:build-{architecture}"
                        for architecture in architectures
                    ]
                },
            },
        )

        self.assertQuerySetEqual(
            merge_uploads.dependencies.all(),
            list(
                WorkRequest.objects.filter(
                    task_name__in={"makesourcepackageupload", "sbuild"}
                )
            ),
            ordered=False,
        )

        self.assertEqual(
            wr.children.filter(task_name="packageupload").count(), 1
        )
        [package_upload] = wr.children.filter(task_name="packageupload")
        self.assertEqual(
            package_upload.task_data,
            {
                "input": {
                    "upload": (
                        f"internal@collections/name:"
                        f"package-upload-merged-{wr.id}"
                    )
                },
                "target": "sftp://upload.example.org/queue",
                "delayed_days": None,
            },
        )

        # Replace one of the promises with an artifact and run the workflow
        # orchestrator again.  The result is stable: no new MergeUploads
        # work request is created.
        binary, _ = self.playground.create_artifact(
            category=ArtifactCategory.UPLOAD,
            data=DebianUpload(
                type="dpkg",
                changes_fields={
                    "Architecture": architectures[0],
                    "Files": [{"name": f"hello_1.0_{architectures[0]}.deb"}],
                },
            ),
        )
        parent.internal_collection.manager.add_artifact(
            binary,
            user=self.playground.get_default_user(),
            workflow=parent,
            name=f"build-{architectures[0]}",
            replace=True,
        )
        self.assertTrue(orchestrate_workflow(wr))

        self.assertEqual(
            wr.children.filter(task_name="mergeuploads").count(), 1
        )

    def test_upload_debsign(self) -> None:
        """Upload signed artifacts."""
        source, [binary], _ = self.create_source_binary_artifacts()

        wr = self.orchestrate(
            data={
                "source_artifact": source.id,
                "binary_artifacts": [binary.id],
                "require_signature": False,
                "target_distribution": "debian/match:codename=bookworm",
                "merge_uploads": True,
                "key": "ABC123",
                "vendor": "debian",
                "codename": "bookworm",
            },
        )

        [signer] = wr.children.filter(
            task_type=TaskTypes.SIGNING, task_name="debsign"
        )
        self.assertEqual(
            signer.task_data,
            {
                "key": "ABC123",
                "unsigned": (
                    f"internal@collections/name:package-upload-merged-{wr.id}"
                ),
            },
        )
        identifier = f"internal@collections/name:package-upload-merged-{wr.id}"
        self.assertEqual(
            signer.workflow_data_json,
            {
                "display_name": f"Sign upload for {identifier}",
                "step": f"debsign-{identifier}",
            },
        )

        self.assertQuerySetEqual(
            signer.dependencies.all(),
            list(
                WorkRequest.objects.filter(
                    task_type=TaskTypes.WORKER, task_name="mergeuploads"
                )
            ),
        )
        signed_name = (
            f"package-upload-signed-"
            f"internal_collections_name_package-upload-merged-{wr.id}"
        )
        self.assert_work_request_event_reactions(
            signer,
            on_success=[
                ActionUpdateCollectionWithArtifacts(
                    artifact_filters={"category": "debian:upload"},
                    collection="internal@collections",
                    created_at=None,
                    name_template=signed_name,
                    variables=None,
                )
            ],
        )

        [uploader] = wr.children.filter(task_name="packageupload")
        self.assertEqual(
            uploader.task_data,
            {
                "input": {"upload": f"internal@collections/name:{signed_name}"},
                "target": "sftp://upload.example.org/queue",
                "delayed_days": None,
            },
        )

        self.assertEqual(
            uploader.workflow_data_json,
            {
                "display_name": (
                    f"Package upload "
                    f"internal@collections/name:package-upload-merged-{wr.id}"
                ),
                "step": (
                    f"package-upload-internal@collections/"
                    f"name:package-upload-merged-{wr.id}"
                ),
            },
        )

    def test_upload_externaldebsign(self) -> None:
        """Upload external signed artifacts."""
        source, [binary], _ = self.create_source_binary_artifacts()

        wr = self.orchestrate(
            data={
                "source_artifact": source.id,
                "binary_artifacts": [binary.id],
                "target_distribution": "debian/match:codename=bookworm",
                "merge_uploads": True,
                "require_signature": True,
                "vendor": "debian",
                "codename": "bookworm",
            },
        )

        [signer] = wr.children.filter(
            task_type=TaskTypes.WAIT, task_name="externaldebsign"
        )
        self.assertEqual(
            signer.task_data,
            {
                "unsigned": (
                    f"internal@collections/name:package-upload-merged-{wr.id}"
                )
            },
        )
        identifier = f"internal@collections/name:package-upload-merged-{wr.id}"
        self.assertEqual(
            signer.workflow_data_json,
            {
                "display_name": (
                    f"Wait for signature on upload for {identifier}"
                ),
                "step": f"external-debsign-{identifier}",
                "needs_input": True,
            },
        )

        self.assertQuerySetEqual(
            signer.dependencies.all(),
            list(
                WorkRequest.objects.filter(
                    task_type=TaskTypes.WORKER, task_name="mergeuploads"
                )
            ),
        )
        signed_name = (
            f"package-upload-signed-"
            f"internal_collections_name_package-upload-merged-{wr.id}"
        )
        self.assert_work_request_event_reactions(
            signer,
            on_success=[
                ActionUpdateCollectionWithArtifacts(
                    artifact_filters={"category": "debian:upload"},
                    collection="internal@collections",
                    created_at=None,
                    name_template=signed_name,
                    variables=None,
                )
            ],
        )

        [uploader] = wr.children.filter(task_name="packageupload")
        self.assertEqual(
            uploader.task_data,
            {
                "input": {"upload": f"internal@collections/name:{signed_name}"},
                "target": "sftp://upload.example.org/queue",
                "delayed_days": None,
            },
        )

    def test_upload_split_signing(self) -> None:
        """Upload artifacts: developer signs source, Debusine signs binary."""
        source, [binary], _ = self.create_source_binary_artifacts()

        wr = self.orchestrate(
            data={
                "source_artifact": source.id,
                "binary_artifacts": [binary.id],
                "require_signature": True,
                "target_distribution": "debian/match:codename=bookworm",
                "binary_key": "ABC123",
                "vendor": "debian",
                "codename": "bookworm",
            },
        )

        [source_signer] = wr.children.filter(
            task_type=TaskTypes.WAIT, task_name="externaldebsign"
        )
        source_lookup = "internal@collections/name:package-upload-source"
        self.assertEqual(source_signer.task_data, {"unsigned": source_lookup})
        self.assertEqual(
            source_signer.workflow_data_json,
            {
                "display_name": (
                    f"Wait for signature on upload for {source_lookup}"
                ),
                "step": f"external-debsign-{source_lookup}",
                "needs_input": True,
            },
        )
        self.assertQuerySetEqual(
            source_signer.dependencies.all(),
            list(
                WorkRequest.objects.filter(
                    task_type=TaskTypes.WORKER,
                    task_name="makesourcepackageupload",
                )
            ),
        )
        source_signed_name = (
            "package-upload-signed-"
            "internal_collections_name_package-upload-source"
        )
        self.assert_work_request_event_reactions(
            source_signer,
            on_success=[
                ActionUpdateCollectionWithArtifacts(
                    artifact_filters={"category": "debian:upload"},
                    collection="internal@collections",
                    name_template=source_signed_name,
                )
            ],
        )

        [binary_signer] = wr.children.filter(
            task_type=TaskTypes.SIGNING, task_name="debsign"
        )
        self.assertEqual(
            binary_signer.task_data,
            {"key": "ABC123", "unsigned": binary.id},
        )
        self.assertEqual(
            binary_signer.workflow_data_json,
            {
                "display_name": f"Sign upload for {binary.id}",
                "step": f"debsign-{binary.id}",
            },
        )
        self.assertQuerySetEqual(
            binary_signer.dependencies.all(), [source_signer]
        )
        binary_signed_name = f"package-upload-signed-{binary.id}"
        self.assert_work_request_event_reactions(
            binary_signer,
            on_success=[
                ActionUpdateCollectionWithArtifacts(
                    artifact_filters={"category": "debian:upload"},
                    collection="internal@collections",
                    created_at=None,
                    name_template=binary_signed_name,
                    variables=None,
                )
            ],
        )

        [source_uploader, binary_uploader] = wr.children.filter(
            task_name="packageupload"
        ).order_by("id")
        self.assertEqual(
            source_uploader.task_data,
            {
                "input": {
                    "upload": f"internal@collections/name:{source_signed_name}"
                },
                "target": "sftp://upload.example.org/queue",
                "delayed_days": None,
            },
        )
        self.assertEqual(
            source_uploader.workflow_data_json,
            {
                "display_name": f"Package upload {source_lookup}",
                "step": f"package-upload-{source_lookup}",
            },
        )
        self.assertQuerySetEqual(
            source_uploader.dependencies.all(), [source_signer]
        )
        self.assertEqual(
            binary_uploader.task_data,
            {
                "input": {
                    "upload": f"internal@collections/name:{binary_signed_name}"
                },
                "target": "sftp://upload.example.org/queue",
                "delayed_days": None,
            },
        )
        self.assertEqual(
            binary_uploader.workflow_data_json,
            {
                "display_name": f"Package upload {binary.id}",
                "step": f"package-upload-{binary.id}",
            },
        )
        self.assertQuerySetEqual(
            binary_uploader.dependencies.all(),
            [binary_signer, source_uploader],
            ordered=False,
        )

    def test_upload_delayed(self) -> None:
        """Upload to a delayed queue."""
        source = self.playground.create_minimal_source_package_artifact()

        wr = self.orchestrate(
            data={
                "source_artifact": source.id,
                "delayed_days": 3,
                "vendor": "debian",
                "codename": "bookworm",
            }
        )

        signed_name = (
            "package-upload-signed-"
            "internal_collections_name_package-upload-source"
        )
        uploader = wr.children.filter(task_name="packageupload").get()
        self.assertEqual(
            uploader.task_data,
            {
                "input": {"upload": f"internal@collections/name:{signed_name}"},
                "target": "sftp://upload.example.org/queue",
                "delayed_days": 3,
            },
        )

    def test_orchestrate_idempotent(self) -> None:
        """Calling orchestrate twice does not create new work requests."""
        self.playground.create_collection(
            "test", CollectionCategory.TASK_CONFIGURATION
        )
        source, _, binaries_collection = self.create_source_binary_artifacts(
            architectures=("amd64", "i386")
        )

        # Force a predictable default ordering.
        with patch(
            "debusine.db.models.collections.CollectionItem._meta.ordering",
            ["id"],
        ):
            wr = self.orchestrate(
                data={
                    "task_configuration": (
                        f"test@{CollectionCategory.TASK_CONFIGURATION}"
                    ),
                    "source_artifact": source.id,
                    "binary_artifacts": {
                        "collection": binaries_collection.id,
                        "name__startswith": "binary-",
                    },
                    "vendor": "debian",
                    "codename": "bookworm",
                    "merge_uploads": True,
                },
            )

        children = set(wr.children.all())

        # Force a different predictable default ordering.
        with patch(
            "debusine.db.models.collections.CollectionItem._meta.ordering",
            ["-id"],
        ):
            self.get_workflow(wr).populate()

        self.assertQuerySetEqual(wr.children.all(), children, ordered=False)

    def test_compute_scheduler_tags(self) -> None:
        self.assertComputesSchedulerTags()

    def test_compute_dynamic_data_source_artifact(self) -> None:
        self.assertEqual(
            self.compute_dynamic_data(self.task),
            PackageUploadWorkflowDynamicData(
                subject="hello",
                parameter_summary="hello → sftp://upload.example.org/queue/",
                source_artifact_id=self.source_artifact.id,
                binary_artifacts_ids=[],
            ),
        )

    def test_compute_dynamic_data_source_is_promise(self) -> None:
        parent = self.playground.create_workflow()
        assert parent.internal_collection is not None

        signing = parent.create_child_worker(
            task_name="noop",
        )
        self.playground.create_bare_data_item(
            parent.internal_collection,
            "source",
            category=BareDataCategory.PROMISE,
            data=DebusinePromise(
                promise_work_request_id=signing.id,
                promise_workflow_id=parent.id,
                promise_category=ArtifactCategory.SOURCE_PACKAGE,
                **{"source_package_name": "hello"},  # noqa: PIE804
            ),
        )

        self.task_data["source_artifact"] = "internal@collections/name:source"
        self.parent = parent

        self.assertEqual(
            self.compute_dynamic_data(self.task),
            PackageUploadWorkflowDynamicData(
                parameter_summary="hello → sftp://upload.example.org/queue/",
                subject="hello",
                binary_artifacts_ids=[],
            ),
        )

    def test_compute_dynamic_data_binary_artifacts(self) -> None:
        binary_artifact = (
            self.playground.create_minimal_binary_package_artifact(
                srcpkg_name="hello"
            )
        )

        del self.task_data["source_artifact"]
        self.task_data["binary_artifacts"] = [binary_artifact.id]
        self.assertEqual(
            self.compute_dynamic_data(self.task),
            PackageUploadWorkflowDynamicData(
                subject="hello",
                parameter_summary="hello → sftp://upload.example.org/queue/",
                source_artifact_id=None,
                binary_artifacts_ids=[binary_artifact.id],
            ),
        )

    def test_get_input_artifacts_source_artifact(self) -> None:
        workflow = self.task
        workflow.dynamic_data = workflow.compute_dynamic_data()
        assert workflow.data.source_artifact is not None
        self.assertEqual(
            workflow.get_input_artifacts(),
            [
                InputArtifactSingle(
                    lookup=workflow.data.source_artifact,
                    label="source_artifact",
                    artifact_id=self.source_artifact.id,
                ),
                InputArtifactMultiple.from_ids(
                    lookup=workflow.data.binary_artifacts,
                    label="binary_artifacts",
                    artifact_ids=[],
                ),
            ],
        )

    def test_get_input_artifacts_binary_artifacts(self) -> None:
        binary_artifact = (
            self.playground.create_minimal_binary_package_artifact(
                srcpkg_name="hello"
            )
        )
        del self.task_data["source_artifact"]
        self.task_data["binary_artifacts"] = [binary_artifact.id]

        workflow = self.task
        workflow.dynamic_data = workflow.compute_dynamic_data()
        self.assertEqual(
            workflow.get_input_artifacts(),
            [
                InputArtifactMultiple.from_ids(
                    lookup=workflow.data.binary_artifacts,
                    label="binary_artifacts",
                    artifact_ids=[binary_artifact.id],
                )
            ],
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Generate cloud-init scripts for bootstrapping Debusine Workers."""

from base64 import b64encode
from textwrap import dedent

import yaml
from django.conf import settings

from debusine.assets.models import APTAuthenticationData
from debusine.db.models.auth import Token
from debusine.server.worker_pools.models import (
    BaseWorkerPoolDataModel,
    DebianRelease,
    DebusineInstallSource,
)
from debusine.tasks.models import ExtraExternalRepository


def apt_source(
    source: DebusineInstallSource,
    custom_source: ExtraExternalRepository | None = None,
    release: DebianRelease | None = None,
) -> ExtraExternalRepository:
    """Generate the APT source for source."""
    if release is None:
        release = DebianRelease.TRIXIE
    match source:
        case DebusineInstallSource.CUSTOM:
            if custom_source is None:
                raise ValueError(
                    "debusine_install_source_custom is not specified"
                )
            return custom_source
        case DebusineInstallSource.DAILY_BUILDS:
            signing_key = dedent(
                """\
                -----BEGIN PGP PUBLIC KEY BLOCK-----
                .
                mDMEaaX1bBYJKwYBBAHaRw8BAQdAfpkAx9/TxpbnGHezlUiruo7YT0EcFHEQ9Swt
                65r8wEa0KUFyY2hpdmUgc2lnbmluZyBrZXkgZm9yIGZyZWV4aWFuL2RlYnVzaW5l
                iJAEExYKADgWIQT8mEBDdxrXXLXVVA1rNLmlWenTHAUCaaX1bAIbAwULCQgHAgYV
                CgkICwIEFgIDAQIeAQIXgAAKCRBrNLmlWenTHDBIAP0S1OwE1zTfV+0jFXYZW6J8
                dcS+fKF1S1YCi6cXLuWzZwEApcVXixlRDzOmQQBymyG35nf2ZRYFuhLOlkD/Ad/W
                Xg8=
                =VWZC
                -----END PGP PUBLIC KEY BLOCK-----
                """
            )
            return ExtraExternalRepository(
                components=["main"],
                url="https://deb.debusine.freexian.com/freexian/debusine",
                suite=f"{release}-devel",
                signing_key=signing_key,
            )
        case DebusineInstallSource.ENTERPRISE:
            suite = f"{release}-enterprise"
        case DebusineInstallSource.ENTERPRISE_STAGING:
            suite = f"{release}-staging"
        case _:  # pragma: no cover
            raise NotImplementedError(f"Unknown source: {source}")

    signing_key = dedent(
        """\
        -----BEGIN PGP PUBLIC KEY BLOCK-----
        .
        mDMEablL9RYJKwYBBAHaRw8BAQdAtjhG7o+/dGAdH1FzsOLmz12j7eUYErbaAodm
        Uhi2tGC0NEFyY2hpdmUgc2lnbmluZyBrZXkgZm9yIGZyZWV4aWFuL2RlYnVzaW5l
        LWVudGVycHJpc2WIkAQTFgoAOBYhBHoYaXCPZVaNbZ2nHZhqj900jkTdBQJpuUv1
        AhsDBQsJCAcCBhUKCQgLAgQWAgMBAh4BAheAAAoJEJhqj900jkTdZ3gA/35jIyT/
        SpM7IJvSta8mX/ntoiGX1Kz0RY+vtAg8ZqfnAQDpEbh76mrYhOH8260HKquubrkJ
        B9EY63RFDSkG3gDbBA==
        =i1MD
        -----END PGP PUBLIC KEY BLOCK-----
        """
    )
    return ExtraExternalRepository(
        components=["main"],
        url="https://deb.debusine.freexian.com/freexian/debusine-enterprise",
        suite=suite,
        signing_key=signing_key,
    )


FIREWALL = dedent(
    """\
    #!/usr/sbin/nft -f

    add table inet filter;
    flush table inet filter;

    table inet filter {
        chain input {
            type filter hook input priority filter;
            iif "lo" accept;  # codespell:ignore iif
            tcp dport ssh counter accept;
            udp dport dhcpv6-client counter accept;
            meta l4proto {icmp, icmpv6} counter accept;
            iifname "debusinebr0" udp dport {domain, bootps} accept;
            ct state invalid,new,untracked counter drop;
        }
        chain forward {
            type filter hook forward priority filter;
            # Cloud Metadata endpoints
            ip daddr 169.254.169.254 counter drop;
            ip6 daddr fd00:ec2::254 counter drop;
        }
        chain output {
            type filter hook output priority filter;
            # Cloud Metadata endpoints
            ip daddr 169.254.169.254 meta skuid != 0 counter drop;
            ip6 daddr fd00:ec2::254 meta skuid != 0 counter drop;
        }
    }
    """
)


class AptConfig(BaseWorkerPoolDataModel):
    """Model for cloud-init APT Configure object."""

    preserve_sources_list: bool
    sources_list: str


class WriteFile(BaseWorkerPoolDataModel):
    """Model for cloud-init Write Files write_files objects."""

    path: str
    content: str
    permissions: str | None = None
    owner: str | None = None


class FilesystemSetup(BaseWorkerPoolDataModel):
    """Model for cloud-init Disk Setup fs_setup objects."""

    device: str
    filesystem: str


class CloudInitConfig(BaseWorkerPoolDataModel):
    """Model for cloud-init configuration."""

    apt: AptConfig | None = None
    device_aliases: dict[str, str] = {}
    fs_setup: list[FilesystemSetup] = []
    hostname: str | None = None
    mounts: list[list[str]] = []
    package_reboot_if_required: bool = False
    package_update: bool = False
    package_upgrade: bool = False
    packages: list[str] = []
    bootcmd: list[str] = []
    runcmd: list[str] = []
    write_files: list[WriteFile] = []

    def as_str(self) -> str:
        """Render configuration file in YAML."""
        return "#cloud-config\n" + yaml.safe_dump(
            self.model_dump(exclude_defaults=True)
        )

    def as_base64_str(self) -> str:
        """Render configuration file in YAML and base64-encode the result."""
        return b64encode(self.as_str().encode("utf-8")).decode("ascii")


def worker_bootstrap_cloud_init(
    worker_name: str,
    activation_token: Token,
    *,
    debusine_install_source: DebusineInstallSource,
    debusine_install_source_custom: ExtraExternalRepository | None = None,
    debian_release: DebianRelease | None = None,
    apt_authentication: APTAuthenticationData | None = None,
) -> CloudInitConfig:
    """Generate cloud-init configuration for a new instance."""
    if getattr(activation_token, "key", None) is None:
        raise ValueError(
            f"Requires a fresh token with a key. Token<{activation_token.pk}> "
            f"does not have one."
        )

    if debusine_install_source == DebusineInstallSource.PRE_INSTALLED:
        return CloudInitConfig(
            package_update=True,
            package_upgrade=True,
            package_reboot_if_required=True,
            hostname=worker_name,
            write_files=[
                WriteFile(
                    path="/etc/debusine/worker/activation-token",
                    permissions="0o600",
                    content=activation_token.key,
                    owner="debusine-worker:debusine-worker",
                ),
            ],
            runcmd=["systemctl restart debusine-worker.service"],
        )

    config = CloudInitConfig(
        package_update=True,
        package_upgrade=True,
        package_reboot_if_required=True,
        apt=None,
        bootcmd=[
            # Executes early in boot, before apt is configured
            (
                "[ -e /usr/bin/eatmydata ] "
                "|| (apt-get update && apt-get -y install eatmydata)"
            ),
            # Don't start the worker until we're ready
            "systemctl mask debusine-worker.service",
        ],
        hostname=worker_name,
        write_files=[
            # We set the owner in a runcmd after the user has been created
            WriteFile(
                path="/etc/debusine/worker/activation-token",
                permissions="0o600",
                content=activation_token.key,
            ),
            WriteFile(
                path="/etc/nftables.conf",
                permissions="0o755",
                content=FIREWALL,
            ),
            WriteFile(
                path="/etc/debusine/worker/config.ini",
                content=dedent(
                    f"""\
                    [General]
                    api-url = https://{settings.DEBUSINE_FQDN}/api
                    log-file = /var/log/debusine/worker/worker.log
                    log-level = DEBUG
                    """
                ),
            ),
        ],
        packages=[
            "autopkgtest",
            "ca-certificates",
            "debusine-worker",
            "gpg",
            "incus",
            "mmdebstrap",
            "nftables",
            "sbuild",
            "uidmap",
        ],
        runcmd=[
            (
                "chown debusine-worker:debusine-worker "
                "/etc/debusine/worker/activation-token"
            ),
            "/usr/share/doc/debusine-worker/examples/configure-worker-incus.sh",
            "systemctl enable nftables.service",
            "systemctl start nftables.service",
            "systemctl unmask debusine-worker.service",
            "systemctl restart debusine-worker.service",
        ],
    )
    backports_packages: list[str] = []
    if debusine_install_source == DebusineInstallSource.BACKPORTS:
        backports_packages.append("src:debusine")

    if backports_packages:
        if debian_release:
            pin_release = f"n={debian_release}"
        else:
            pin_release = "a=stable"
        config.write_files += [
            WriteFile(
                path="/etc/apt/preferences.d/90-debusine-backports",
                content=dedent(
                    f"""\
                    Package: {" ".join(backports_packages)}
                    Pin: release {pin_release}-backports
                    Pin-Priority: 900

                    Package: {" ".join(backports_packages)}
                    Pin: release {pin_release}
                    Pin-Priority: -1
                    """
                ),
            ),
        ]
    if debusine_install_source in (
        DebusineInstallSource.DAILY_BUILDS,
        DebusineInstallSource.ENTERPRISE,
        DebusineInstallSource.ENTERPRISE_STAGING,
        DebusineInstallSource.CUSTOM,
    ):
        source = apt_source(
            debusine_install_source,
            debusine_install_source_custom,
            debian_release,
        )
        config.write_files += [
            WriteFile(
                path="/etc/apt/sources.list.d/debusine.sources",
                content=source.as_deb822_source(),
            ),
            WriteFile(
                path="/etc/apt/preferences.d/90-debusine-ignore-debian",
                content=dedent(
                    """\
                    Package: src:debusine
                    Pin: release o=Debian Backports
                    Pin-Priority: -1

                    Package: src:debusine
                    Pin: release o=Debian
                    Pin-Priority: -1
                    """
                ),
            ),
        ]
        if apt_authentication:
            config.write_files += [
                WriteFile(
                    path=f"/etc/apt/auth.conf.d/{apt_authentication.name}.conf",
                    content=dedent(
                        f"""\
                        machine {source.url}
                        login {apt_authentication.credentials.username}
                        password {apt_authentication.credentials.password}
                        """
                    ),
                ),
            ]
        elif debusine_install_source in (
            DebusineInstallSource.ENTERPRISE,
            DebusineInstallSource.ENTERPRISE_STAGING,
        ):
            raise ValueError(
                "apt_authentication is required for ENTERPRISE sources."
            )

    return config



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests the scopes UI helper."""

from django.core.exceptions import ImproperlyConfigured
from django.test import RequestFactory
from django.urls import reverse

from debusine.db.context import context
from debusine.db.permissioncontext import PermissionContext
from debusine.db.playground import scenarios
from debusine.test.django import TestCase
from debusine.web.icons import Icons
from debusine.web.views.places import ContainerPlace
from debusine.web.views.ui.scopes import ScopePlace, ScopeUI


class ScopeUITests(TestCase):
    """Test the :py:class:`ScopeUI` class."""

    scenario = scenarios.DefaultContext(set_current=True)

    def ui(self) -> ScopeUI:
        """Instantiate the UI helper to test."""
        assert context.user
        factory = RequestFactory()
        request = factory.get("/")
        request.user = context.user
        return self.scenario.scope.ui(request)

    def test_place(self) -> None:
        place = self.ui().place
        self.assertEqual(place.title, self.scenario.scope.name)
        self.assertEqual(place.description, "debusine scope")
        self.assertEqual(place.url, reverse("scopes:detail"))
        self.assertEqual(place.icon, Icons.SCOPE)
        self.assertIsNone(place.parent)
        self.assertEqual(place.breadcrumb, self.scenario.scope.name)

    def test_place_with_icon(self) -> None:
        self.scenario.scope.icon = "icon.png"
        self.scenario.scope.save()
        place = self.ui().place
        self.assertIsInstance(place, ScopePlace)
        self.assertEqual(place.title, self.scenario.scope.name)
        self.assertEqual(place.description, "debusine scope")
        self.assertEqual(place.url, reverse("scopes:detail"))
        self.assertEqual(place.icon, "icon.png")
        self.assertIsNone(place.parent)
        self.assertEqual(place.breadcrumb, self.scenario.scope.name)

    def test_place_group_list(self) -> None:
        place = self.ui().place_group_list
        self.assertIsInstance(place, ContainerPlace)
        self.assertEqual(
            place.title, "Group list for playground in scope debusine"
        )
        self.assertEqual(
            place.description,
            f"Groups in scope debusine accessible by {self.scenario.user}",
        )
        self.assertEqual(
            place.url, reverse("user:groups", kwargs={"username": "playground"})
        )
        self.assertIsNone(place.icon)
        self.assertEqual(place.breadcrumb, "groups")
        assert place.parent is not None
        self.assertEqual(place.parent.url, reverse("scopes:detail"))

    @context.local()
    def test_place_group_list_unauthenticated(self) -> None:
        context.reset()
        context.set_permission_context_scoped(
            self.scenario.scope, PermissionContext.for_ui_anonymous()
        )
        with self.assertRaisesRegex(
            ImproperlyConfigured,
            "ScopeUI.place_group_list invoked by an anonymous user",
        ):
            self.ui().place_group_list


class ScopePlaceTests(TestCase):
    """Test the :py:class:`ScopePlace` class."""

    def test_as_nav(self) -> None:
        place = ScopePlace(
            title="test title", url="/test-url", breadcrumb="test"
        )
        tree = self.assertHTMLValid(place.as_nav())
        a = self.assertHasElement(tree, "//a")
        self.assertTextContentEqual(a, "test title")
        self.assertEqual(a.get("href"), "/test-url")

    def test_as_nav_icon(self) -> None:
        place = ScopePlace(
            title="test title",
            url="/test-url",
            icon="icon.png",
            description="description",
            breadcrumb="test",
        )
        tree = self.assertHTMLValid(place.as_nav())
        a = self.assertHasElement(tree, "//a")
        self.assertTextContentEqual(a, "test title")
        self.assertEqual(a.get("href"), "/test-url")

        img = self.assertHasElement(a, "img")
        self.assertEqual(img.get("src"), "/static/icon.png")

    def test_as_page_title(self) -> None:
        place = ScopePlace(title="test title", url="url", breadcrumb="test")
        tree = self.assertHTMLValid(place.as_page_title())
        title = self.assertHasElement(tree, "//h1")
        self.assertTextContentEqual(title, "test title")

    def test_as_page_title_icon(self) -> None:
        place = ScopePlace(
            title="test title", url="url", icon="icon.png", breadcrumb="test"
        )
        tree = self.assertHTMLValid(place.as_page_title())
        h1 = self.assertHasElement(tree, "//h1")
        self.assertTextContentEqual(h1, "test title")

    def test_as_page_title_icon_title(self) -> None:
        place = ScopePlace(
            title="test title",
            url="url",
            icon="icon.png",
            description="distinction",
            breadcrumb="test",
        )
        tree = self.assertHTMLValid(place.as_page_title())
        h1 = self.assertHasElement(tree, "//h1")
        self.assertTextContentEqual(h1, "test title")



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests related to workers views."""

import datetime as dt
from typing import ClassVar, override

from django.urls import reverse
from rest_framework import status

from debusine.db.models import WorkRequest, Worker
from debusine.db.playground import scenarios
from debusine.tasks.models import WorkerType
from debusine.test.django import (
    AllowAll,
    DenyAll,
    TestCase,
    TestResponseType,
    override_permission,
)
from debusine.web.icons import Icons
from debusine.web.views.tests.utils import ViewTestMixin


class WorkersListViewTests(ViewTestMixin, TestCase):
    """Tests for :py:class:`WorkersListView`."""

    scenario = scenarios.DefaultContext(set_current=True)

    def test_template_with_workers(self) -> None:
        """Test template contains expected information."""
        worker_external = self.playground.create_worker()
        worker_external.name = "a-first-one"
        worker_external.concurrency = 3
        worker_external.save()
        worker_external.mark_connected()

        worker_celery = self.playground.create_worker(
            worker_type=WorkerType.CELERY
        )
        worker_celery.name = "b-second-one"
        worker_celery.save()

        worker_signing = self.playground.create_worker(
            worker_type=WorkerType.SIGNING
        )
        worker_signing.name = "c-third-one"
        worker_signing.save()

        work_request_1 = self.playground.create_worker_task(
            task_name="noop",
            worker=worker_external,
            mark_running=True,
            workspace=self.scenario.workspace,
        )
        work_request_2 = self.playground.create_worker_task(
            task_name="noop",
            worker=worker_external,
            mark_running=True,
            workspace=self.scenario.workspace,
        )

        response = self.client.get(reverse("workers:list"))
        tree = self.assertResponseHTML(response)
        table = self.assertHasElement(tree, "//table[@id='worker-list']")

        # Assert names
        self.assertEqual(len(table.tbody.tr), 2)
        self.assertEqual(table.tbody.tr[0].td[0].i.get("title"), "External")
        self.assertEqual(
            table.tbody.tr[0].td[0].i.get("class"),
            f"bi bi-{Icons.WORKER_EXTERNAL}",
        )
        self.assertEqual(table.tbody.tr[1].td[0].i.get("title"), "Signing")
        self.assertEqual(
            table.tbody.tr[1].td[0].i.get("class"),
            f"bi bi-{Icons.WORKER_SIGNING}",
        )

        # Assert status for external worker: "Running ..."
        path_1 = work_request_1.get_absolute_url()
        path_2 = work_request_2.get_absolute_url()
        external_running = (
            f'<td> '
            f'<a href="{path_1}">noop</a><br/>'
            f'<a href="{path_2}">noop</a><br/>'
            f'</td>'
        )
        self.assertHTMLContentsEquivalent(
            table.tbody.tr[0].td[4], external_running
        )

    def test_status_disabled(self) -> None:
        """View shows "Disabled" (worker's token is disabled)."""
        worker = self.playground.create_worker()
        assert worker.token is not None
        worker.token.enabled = False
        worker.token.save()

        response = self.client.get(reverse("workers:list"))
        tree = self.assertResponseHTML(response)
        table = self.assertHasElement(tree, "//table[@id='worker-list']")

        self.assertHTMLContentsEquivalent(
            table.tbody.tr[0].td[4],
            '<td><span class="badge text-bg-danger">Disabled</span></td>',
        )

    def test_status_idle(self) -> None:
        """View shows "Idle" (enabled worker, not running any work request)."""
        worker = self.playground.create_worker()
        worker.mark_connected()

        response = self.client.get(reverse("workers:list"))
        tree = self.assertResponseHTML(response)
        table = self.assertHasElement(tree, "//table[@id='worker-list']")

        self.assertHTMLContentsEquivalent(
            table.tbody.tr[0].td[4],
            '<td> <span class="badge text-bg-success">Idle</span> </td>',
        )

    def test_status_disconnected(self) -> None:
        """View shows "Disconnected" (enabled worker, not connected)."""
        self.playground.create_worker()

        response = self.client.get(reverse("workers:list"))
        tree = self.assertResponseHTML(response)
        table = self.assertHasElement(tree, "//table[@id='worker-list']")

        self.assertHTMLContentsEquivalent(
            table.tbody.tr[0].td[4],
            '<td> <span class="badge text-bg-warning">'
            'Disconnected</span> </td>',
        )

    def test_template_no_workers(self) -> None:
        """Test template renders "No workers registered in this..."."""
        response = self.client.get(reverse("workers:list"))

        self.assertContains(
            response,
            "<p>No workers registered in this debusine instance.</p>",
            html=True,
        )

    def test_template_with_private_work_request_anonymous(self) -> None:
        """Test template hides private work requests from anonymous viewers."""
        self.scenario.workspace.public = False
        self.scenario.workspace.save()

        worker_external = self.playground.create_worker()
        worker_external.name = "a-first-one"
        worker_external.concurrency = 3
        worker_external.save()
        worker_external.mark_connected()

        self.playground.create_worker_task(
            task_name="noop",
            worker=worker_external,
            mark_running=True,
            workspace=self.scenario.workspace,
        )

        response = self.client.get(reverse("workers:list"))
        tree = self.assertResponseHTML(response)
        table = self.assertHasElement(tree, "//table[@id='worker-list']")

        # Assert status for external worker: "Running ..."
        external_running = '<td>Private Task<br/></td>'
        self.assertHTMLContentsEquivalent(
            table.tbody.tr[0].td[4], external_running
        )

    def test_template_with_private_work_request_authenticated(self) -> None:
        """Test template shows private work requests to the authenticated."""
        worker_external = self.playground.create_worker()
        worker_external.name = "a-first-one"
        worker_external.concurrency = 3
        worker_external.save()
        worker_external.mark_connected()

        work_request = self.playground.create_worker_task(
            task_name="noop",
            worker=worker_external,
            mark_running=True,
            workspace=self.scenario.workspace,
        )

        self.client.force_login(self.scenario.user)
        response = self.client.get(reverse("workers:list"))
        tree = self.assertResponseHTML(response)
        table = self.assertHasElement(tree, "//table[@id='worker-list']")

        # Assert status for external worker: "Running ..."
        path = work_request.get_absolute_url()
        external_running = f'<td><a href="{path}">noop</a><br/></td>'
        self.assertHTMLContentsEquivalent(
            table.tbody.tr[0].td[4], external_running
        )


class WorkersListViewOrderingTests(ViewTestMixin, TestCase):
    """Tests for :py:class:`WorkersListView` ordering."""

    scenario = scenarios.DefaultContext(set_current=True)
    worker_external: ClassVar[Worker]
    worker_celery: ClassVar[Worker]
    worker_signing: ClassVar[Worker]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.worker_external = cls.playground.create_worker()
        cls.worker_external.name = "a-external"
        cls.worker_external.concurrency = 3
        cls.worker_external.save()
        cls.worker_external.mark_connected()

        cls.worker_celery = cls.playground.create_worker(
            worker_type=WorkerType.CELERY
        )
        cls.worker_celery.name = "b-celery"
        cls.worker_celery.save()

        cls.worker_signing = cls.playground.create_worker(
            worker_type=WorkerType.SIGNING
        )
        cls.worker_signing.name = "c-signing"
        cls.worker_signing.save()

    def assertTableEntries(
        self, response: TestResponseType, expected: list[str]
    ) -> None:
        """Check that the workers in the table match the given list of names."""
        self.assertEqual(response.status_code, status.HTTP_200_OK)
        tree = self.assertResponseHTML(response)
        table = self.assertHasElement(tree, "//table[@id='worker-list']")
        actual = [
            self.get_node_text_normalized(tr.td[1]) for tr in table.tbody.tr
        ]
        self.assertEqual(actual, expected)

    def test_order_ascending(self) -> None:
        """Test ordering alphabetically."""
        response = self.client.get(reverse("workers:list"), {"order": "name"})
        self.assertTableEntries(response, ["a-external", "c-signing"])

    def test_order_descending(self) -> None:
        """Test ordering reverse-alphabetically."""
        response = self.client.get(
            reverse("workers:list"), {"order": "name.desc"}
        )
        self.assertTableEntries(response, ["c-signing", "a-external"])


class WorkerDetailViewTests(ViewTestMixin, TestCase):
    """Tests for :py:class:`WorkerDetailView`."""

    scenario = scenarios.DefaultContext(set_current=True)

    def test_authenticated(self) -> None:
        """Test a GET with an authenticated user."""
        worker = self.playground.create_worker()
        self.client.force_login(self.scenario.user)
        response = self.client.get(worker.get_absolute_url())
        tree = self.assertResponseHTML(response)

        el = self.assertHasElement(tree, "//tr[@id='row-name']")
        self.assertTextContentEqual(el.td[0], worker.name)
        el = self.assertHasElement(tree, "//tr[@id='row-worker_type']")
        self.assertTextContentEqual(el.td[0], worker.worker_type)
        el = self.assertHasElement(tree, "//tr[@id='row-registered_at']")
        self.assertDatetimeContentEqual(
            el.td[0], worker.registered_at, anywhere=True
        )
        el = self.assertHasElement(tree, "//tr[@id='row-connected_at']")
        self.assertTextContentEqual(el.td[0], "Never")
        el = self.assertHasElement(tree, "//tr[@id='row-instance_created_at']")
        self.assertTextContentEqual(el.td[0], "Never")
        el = self.assertHasElement(tree, "//tr[@id='row-concurrency']")
        self.assertTextContentEqual(el.td[0], str(worker.concurrency))

        el = self.assertHasElement(tree, "//tr[@id='row-metadata']")
        body = self.assertHasElement(
            el, ".//div[contains(@class, 'card-body')]"
        )
        self.assertYAMLContentEqual(body, worker.metadata())
        footer = self.assertHasElement(
            el, ".//div[contains(@class, 'card-footer')]"
        )
        assert worker.dynamic_metadata_updated_at is not None
        self.assertDatetimeContentEqual(
            footer.xpath("//em")[0],
            worker.dynamic_metadata_updated_at,
            anywhere=True,
        )

        self.assertFalse(tree.xpath("//tr[@id='row-worker_pool']"))
        self.assertFalse(tree.xpath("//table[@id='work_request-list']"))

    def test_dates(self) -> None:
        """Test a GET with all dates present."""
        worker = self.playground.create_worker()
        worker.instance_created_at = dt.datetime(
            2024, 1, 1, 0, 0, 0, tzinfo=dt.UTC
        )
        worker.dynamic_metadata_updated_at = dt.datetime(
            2024, 2, 2, 0, 0, 0, tzinfo=dt.UTC
        )
        worker.mark_connected()
        worker.save()
        assert worker.connected_at is not None
        self.client.force_login(self.scenario.user)
        response = self.client.get(worker.get_absolute_url())
        tree = self.assertResponseHTML(response)

        el = self.assertHasElement(tree, "//tr[@id='row-registered_at']")
        self.assertDatetimeContentEqual(
            el.td[0], worker.registered_at, anywhere=True
        )

        el = self.assertHasElement(tree, "//tr[@id='row-connected_at']")
        self.assertDatetimeContentEqual(
            el.td[0], worker.connected_at, anywhere=True
        )

        el = self.assertHasElement(tree, "//tr[@id='row-instance_created_at']")
        self.assertDatetimeContentEqual(
            el.td[0], worker.instance_created_at, anywhere=True
        )

        el = self.assertHasElement(tree, "//tr[@id='row-metadata']")
        footer = self.assertHasElement(
            el, ".//div[contains(@class, 'card-footer')]"
        )
        self.assertDatetimeContentEqual(
            footer.xpath("//em")[0],
            worker.dynamic_metadata_updated_at,
            anywhere=True,
        )

    def test_no_dynamic_metadata_updated_at(self) -> None:
        """Test a GET with no dynamic_metadata_updated_at."""
        worker = self.playground.create_worker()
        worker.dynamic_metadata_updated_at = None
        worker.save()
        self.client.force_login(self.scenario.user)
        response = self.client.get(worker.get_absolute_url())
        tree = self.assertResponseHTML(response)
        el = self.assertHasElement(tree, "//tr[@id='row-metadata']")
        footer = self.assertHasElement(
            el, ".//div[contains(@class, 'card-footer')]"
        )
        self.assertTextContentEqual(footer, "Dynamic metadata never updated.")

    def test_worker_pool(self) -> None:
        """Check presentation of worker pool information."""
        worker_pool = self.playground.create_worker_pool()
        worker = self.playground.create_worker(worker_pool=worker_pool)
        worker.worker_pool_data = {"foo": "bar"}
        worker.save()
        assert worker.worker_pool is not None
        self.client.force_login(self.scenario.user)
        response = self.client.get(worker.get_absolute_url())
        tree = self.assertResponseHTML(response)
        td = self.assertHasElement(tree, "//tr[@id='row-worker_pool']/td")
        card = self.assertHasElement(td, "div[@class='card']")
        self.assertTextContentEqual(card.div[0], worker.worker_pool.name)
        self.assertYAMLContentEqual(card.div[1], worker.worker_pool_data)

    def test_worker_pool_no_data(self) -> None:
        """Check worker pool information without worker_pool_data."""
        worker_pool = self.playground.create_worker_pool()
        worker = self.playground.create_worker(worker_pool=worker_pool)
        worker.worker_pool_data = None
        worker.save()
        assert worker.worker_pool is not None
        self.client.force_login(self.scenario.user)
        response = self.client.get(worker.get_absolute_url())
        tree = self.assertResponseHTML(response)
        td = self.assertHasElement(tree, "//tr[@id='row-worker_pool']/td")
        self.assertFalse(td.xpath("div[@class='card']"))
        self.assertTextContentEqual(td, worker.worker_pool.name)

    def test_worker_pool_anonymous(self) -> None:
        """Anonymous users cannot see worker_pool_data."""
        worker_pool = self.playground.create_worker_pool()
        worker = self.playground.create_worker(worker_pool=worker_pool)
        worker.worker_pool_data = {"foo": "bar"}
        worker.save()
        assert worker.worker_pool is not None
        response = self.client.get(worker.get_absolute_url())
        tree = self.assertResponseHTML(response)
        td = self.assertHasElement(tree, "//tr[@id='row-worker_pool']/td")
        self.assertFalse(td.xpath("div[@class='card']"))
        self.assertTextContentEqual(td, worker.worker_pool.name)

    def test_work_requests_visible(self) -> None:
        """Check the list of assigned work requests."""
        worker = self.playground.create_worker()
        self.playground.create_worker_task(worker=worker)

        self.client.force_login(self.scenario.user)
        with override_permission(WorkRequest, "can_display", AllowAll):
            response = self.client.get(worker.get_absolute_url())
        tree = self.assertResponseHTML(response)
        table = self.assertHasElement(tree, "//table[@id='work_request-list']")
        self.assertEqual(len(table.tbody.tr), 1)

    def test_work_requests_private(self) -> None:
        """Check that private work requests are hidden."""
        self.scenario.workspace.public = False
        self.scenario.workspace.save()
        worker = self.playground.create_worker()
        self.playground.create_worker_task(worker=worker)

        self.client.force_login(self.scenario.user)
        with override_permission(WorkRequest, "can_display", DenyAll):
            response = self.client.get(worker.get_absolute_url())
        tree = self.assertResponseHTML(response)
        self.assertFalse(tree.xpath("//table[@id='work_request-list']"))



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""
Backend infrastructure for provisioning cloud workers.

See docs/reference/devel-blueprints/dynamic-compute.rst
"""

import logging
import math
from collections import defaultdict
from functools import cached_property

from celery import shared_task
from django.contrib.postgres.functions import TransactionNow
from django.db import transaction
from django.db.models import Count, F, IntegerField, Q, Sum
from django.db.models.functions import Cast, Coalesce, Extract, Greatest, Least
from django.utils import timezone

from debusine.db.context import context
from debusine.db.models import (
    Scope,
    ScopeWorkerPool,
    WorkRequest,
    Worker,
    WorkerPool,
)
from debusine.db.models.worker_pools import (
    WorkerPoolStatistics,
    WorkerPoolTaskExecutionStatistics,
)

# TODO: Hardcoded estimation of duration for a task, until we have code to
#       estimate the duration of future tasks based on task-statistics
TASK_DURATION_SECONDS = 10 * 60

# TODO: Hardcoded target latency for a scope, until we can read it from Scope
TARGET_LATENCY_SECONDS = 3600

log = logging.getLogger(__name__)


def terminate_idle(minutes: int = 10) -> bool:
    """
    Find and terminate idle workers.

    :returns: true if idle workers were found
    """
    # List cloud workers that have been idle for more than 10 minutes
    idle_workers = (
        Worker.objects.active()
        .with_idle_time()
        .filter(
            idle_time__gt=minutes * 60,
            worker_pool__isnull=False,
            worker_pool__enabled=True,
        )
    )
    has_idle = False
    for worker in idle_workers:
        assert worker.worker_pool is not None
        has_idle = True
        log.info(
            "Terminating idle worker %s (idle for %d seconds)",
            worker.name,
            worker.idle_time,
        )
        worker.worker_pool.terminate_worker(worker)
    return has_idle


class WorkerAllocator:
    """Allocate new workers needed to handle the pending load."""

    # NOTE: This is a minimum viable implementation of cloud compute
    # provisioning, intended to be useful until we implement tagged scheduling
    # (#326).
    #
    # The following simplifications are in place:
    # * assume a hardcoded running time for tasks, until we have code to
    #   estimate the duration of future tasks based on task-statistics
    # * assume all static workers can handle all tasks
    # * assume all workers take the same time to process a task
    # * limit cloud provisioning to amd64 workers

    def __init__(self) -> None:
        """Prefetch common information from the database."""
        self.start_of_month = timezone.now().replace(
            day=1, hour=0, minute=0, second=0, microsecond=0
        )
        self.static_workers = self.static_workers_count()

    def static_workers_count(self) -> dict[str, int]:
        """Count connected static workers that can build architectures."""
        workers: dict[str, int] = defaultdict(int)
        for row in (
            Worker.objects.connected()
            .filter(worker_pool__isnull=True)
            .annotate(
                architectures=Coalesce(
                    "static_metadata__system:architectures",
                    "dynamic_metadata__system:architectures",
                )
            )
            .values("architectures")
            .annotate(count=Count("id"))
            .order_by()
        ):
            for architecture in row["architectures"]:
                workers[architecture] += row["count"]
        return workers

    def limit_by_max_active_instances(
        self, pool: WorkerPool, amount: int
    ) -> int:
        """Limit amount by the max_active_instances configured in the pool."""
        pool_limits = pool.limits_model
        if pool_limits.max_active_instances is None:
            return amount

        current_worker_count = pool.workers_running.count()
        available = pool_limits.max_active_instances - current_worker_count
        if available <= 0:
            return 0
        return min(amount, available)

    def get_used_seconds(self, pool: WorkerPool) -> int:
        """Count the seconds of pool usage for this month."""
        completed = (
            WorkerPoolStatistics.objects.filter(
                worker_pool=pool, timestamp__gte=self.start_of_month
            )
            .annotate(
                runtime_this_month=Least(
                    "runtime",
                    Cast(
                        Extract(
                            F("timestamp") - self.start_of_month,  # type: ignore[operator] # noqa: E501
                            "epoch",
                        ),
                        IntegerField(),
                    ),
                )
            )
            .aggregate(seconds_per_month=Sum("runtime_this_month"))[
                "seconds_per_month"
            ]
            or 0
        )
        connected = (
            Worker.objects.connected()
            .filter(worker_pool=pool)
            .annotate(
                runtime=Extract(
                    TransactionNow()
                    - Greatest("instance_created_at", self.start_of_month),
                    "epoch",
                )
            )
            .aggregate(seconds_per_month=Sum("runtime"))["seconds_per_month"]
            or 0
        )
        return completed + connected

    def limit_by_pool_target_max_seconds_per_month(
        self, pool: WorkerPool, amount: int
    ) -> int:
        """Limit amount by the pool's target_max_seconds_per_month."""
        target_max_seconds_per_month: int | None = (
            pool.limits_model.target_max_seconds_per_month
        )
        if target_max_seconds_per_month is None:
            return amount

        used_seconds = self.get_used_seconds(pool)
        if used_seconds >= target_max_seconds_per_month:
            return 0

        return min(
            amount,
            int(
                (target_max_seconds_per_month - used_seconds)
                / TASK_DURATION_SECONDS
            ),
        )

    def process_scope(self, scope: Scope) -> None:
        """Perform allocation for a scope."""
        allocator = ScopeWorkerAllocator(self, scope)
        allocator.process_scope()


class ScopeWorkerAllocator:
    """Allocate new workers needed to handle the pending load of a scope."""

    def __init__(self, worker_allocator: WorkerAllocator, scope: Scope) -> None:
        """Prefetch common information from the database."""
        self.worker_allocator = worker_allocator
        self.scope = scope

    def get_used_seconds(self, pool: WorkerPool) -> int:
        """Count the seconds of pool usage for the current scope this month."""
        completed = (
            WorkerPoolTaskExecutionStatistics.objects.filter(
                worker_pool=pool,
                scope=self.scope,
                timestamp__gte=self.worker_allocator.start_of_month,
            )
            .annotate(
                runtime_this_month=Least(
                    "runtime",
                    Cast(
                        Extract(
                            F("timestamp")
                            - self.worker_allocator.start_of_month,  # type: ignore[operator] # noqa: E501
                            "epoch",
                        ),
                        IntegerField(),
                    ),
                )
            )
            .aggregate(seconds_per_month=Sum("runtime_this_month"))[
                "seconds_per_month"
            ]
            or 0
        )
        running = (
            WorkRequest.objects.running()
            .filter(
                worker__worker_pool=pool,
                workspace__scope=self.scope,
                started_at__isnull=False,
            )
            .annotate(
                runtime=Extract(
                    TransactionNow()
                    - Greatest(
                        "started_at", self.worker_allocator.start_of_month
                    ),
                    "epoch",
                )
            )
            .aggregate(seconds_per_month=Sum("runtime"))["seconds_per_month"]
            or 0
        )
        return completed + running

    def limit_by_scope_pool_target_max_seconds_per_month(
        self, pool: WorkerPool, amount: int
    ) -> int:
        """Limit amount by target_max_seconds_per_month for scope and pool."""
        target_max_seconds_per_month: int | None = None
        try:
            scope_worker_pool = ScopeWorkerPool.objects.get(
                worker_pool=pool, scope=self.scope
            )
        except ScopeWorkerPool.DoesNotExist:
            return amount

        target_max_seconds_per_month = (
            scope_worker_pool.limits_model.target_max_seconds_per_month
        )
        if target_max_seconds_per_month is None:
            return amount

        used_seconds = self.get_used_seconds(pool)
        if used_seconds >= target_max_seconds_per_month:
            return 0

        return min(
            amount,
            int(
                (target_max_seconds_per_month - used_seconds)
                / TASK_DURATION_SECONDS
            ),
        )

    def allocate_workers(self, pool: WorkerPool, amount: int) -> int:
        """
        Allocate workers on pool for the given scope.

        The number of workers allocated can be less than amount, or even zero,
        depending on pool availability and quotas.

        :param scope: the scope requesting allocation
        :param pool: the pool used to allocate
        :param amount: the amount of new workers required
        :returns: the amount of workers allocated
        """
        # Enforce max_active_instances
        amount = self.worker_allocator.limit_by_max_active_instances(
            pool, amount
        )

        # Enforce target_max_seconds_per_month
        amount = (
            self.worker_allocator.limit_by_pool_target_max_seconds_per_month(
                pool, amount
            )
        )

        # Enforce target_max_seconds_per_month
        amount = self.limit_by_scope_pool_target_max_seconds_per_month(
            pool, amount
        )

        if amount > 0:
            log.info("Launching %d workers in %s", amount, pool.name)
            pool.launch_workers(amount)
        return amount

    def pending_task_count(self) -> dict[str, int]:
        """Return the number of pending tasks by architecture."""
        pending: dict[str, int] = {}
        for row in (
            WorkRequest.objects.pending()
            .filter(
                workspace__scope=self.scope,
            )
            .values('task_data__build_architecture')
            .annotate(count=Count('id'))
            .order_by()
        ):
            pending[row["task_data__build_architecture"]] = row["count"]
        return pending

    def workers_required(self) -> dict[str, int]:
        """
        Compute how many workers can process the scope's pending load.

        This computes the absolute number of workers required to process the
        scope's pending load for the given architectures, within the expected
        latency targets.
        """
        # If we have no pending tasks, we do not require workers
        pending_tasks = self.pending_task_count()

        workers: dict[str, int] = {}
        for arch, tasks in pending_tasks.items():
            if tasks == 0:
                workers[arch] = 0
                continue

            # Estimated number of worker-seconds needed to run all but the last
            # pending task.
            pending_seconds = TASK_DURATION_SECONDS * max(tasks - 1, 0)

            # Each worker will contribute one second per second of work towards
            # draining the queue.  There's no point in allocating more workers
            # than the number of pending tasks.
            workers[arch] = max(
                1,
                min(math.ceil(pending_seconds / TARGET_LATENCY_SECONDS), tasks),
            )
        return workers

    def workers_available(self) -> dict[str, int]:
        """Compute the number of workers currently available to the scope."""
        workers = defaultdict(int, self.worker_allocator.static_workers)

        for row in (
            Worker.objects.active()
            .filter(
                Q(worker_pool__instance_wide=True)
                | Q(worker_pool__scopeworkerpool__scope=self.scope)
            )
            .values("worker_pool__architectures")
            .annotate(count=Count("id"))
            .order_by()
        ):
            for architecture in row["worker_pool__architectures"]:
                workers[architecture] += row["count"]
        return workers

    @cached_property
    def allocation_pools(self) -> list[WorkerPool]:
        """
        Return the list of pools that can be used to allocate workers.

        The list is ordered by priority, so that allocation can proceed from
        the first element onwards.
        """
        # Pools in which we can allocate them
        pools: list[WorkerPool] = []

        # Add scope pools by decreasing priority
        pools.extend(
            WorkerPool.objects.enabled()
            .filter(scopeworkerpool__scope=self.scope)
            .order_by("-scopeworkerpool__priority")
        )

        # Add instance-wide pools at the end
        pools.extend(WorkerPool.objects.enabled().filter(instance_wide=True))

        # TODO: enforce that ScopeWorkerPool cannot link to instance-wide
        # pools? Otherwise we may need to remove duplicates here

        return pools

    def process_scope(self) -> None:
        """Perform allocation for this scope."""
        # Number of workers required to process the queue in time
        workers_required_by_arch = self.workers_required()
        workers_available_by_arch = self.workers_available()

        # Allocate from one pool at a time
        for pool in self.allocation_pools:
            for architecture in pool.architectures:
                workers_available = workers_available_by_arch.get(
                    architecture, 0
                )
                workers_required = workers_required_by_arch.get(architecture, 0)

                # Number of workers that we need to allocate
                new_workers_required = workers_required - workers_available

                # If we have enough workers already, we are done
                if new_workers_required <= 0:
                    continue

                allocated = self.allocate_workers(pool, new_workers_required)
                workers_required_by_arch[architecture] -= allocated


def create_new_workers() -> None:
    """Provision new workers if needed to handle pending load."""
    # NOTE: This is a minimum viable implementation of cloud compute
    # provisioning, intended to be useful until we implement tagged scheduling
    # (#326).
    #
    # The following simplifications are in place:
    # * assume all workers take the same time to process a task
    # * assume scopes' load demands are isolated from each other, which allows
    #   to process each scope's work queue without taking other scopes into
    #   account.
    #   This is not correct with regards to static workers, instance-wide pools
    #   or pools shared by multiple scopes, but in the worst case we
    #   underestimate the number of new workers needed and end up provisioning
    #   some workers again later as queues drain more slowly than estimated.
    #
    #   i.e. if two scopes would each need 15 workers to meet targets and there
    #   are 10 static workers, they'd each provision 5 from pools in this MVP,
    #   while in reality they'd need more because they're both sharing the same
    #   static workers.
    # * see WorkerAllocator for more assumptions

    allocator = WorkerAllocator()
    for scope in Scope.objects.all():
        allocator.process_scope(scope)


def scrub_workers() -> None:
    """Scrub all WorkerPools."""
    for pool in WorkerPool.objects.all():
        pool.provider_interface.terminate_missing()


# mypy complains that celery.shared_task is untyped, which is true, but we
# can't fix that here.
@shared_task()  # type: ignore[untyped-decorator]
@transaction.atomic
@context.local()
@context.disable_permission_checks()
def provision() -> None:
    """Provision or terminate workers from pools to handle pending tasks."""
    # NOTE: This is a minimum viable implementation of cloud compute
    # provisioning, intended to be useful until we implement tagged scheduling
    # (#326).
    #
    # The following simplifications are in place:
    #  * do not consider per-scope quotas when terminating workers: only idle
    #    workers are terminated
    #  * see create_new_workers() comments for more assumptions

    if terminate_idle():
        # If workers have been idle for more than 10 minutes, we can assume
        # that we do not have a significant backlog of pending work requests.
        #
        # This may not be true: we may have a pool of idle arm64 workers and a
        # backlog of amd64 tasks. If that is the case, we will catch up with
        # the backlog in the next provision run.
        return

    create_new_workers()


# mypy complains that celery.shared_task is untyped, which is true, but we
# can't fix that here.
@shared_task()  # type: ignore[untyped-decorator]
@transaction.atomic
@context.local()
@context.disable_permission_checks()
def scrub() -> None:
    """
    Scan the worker pool for failed workers and clean them up.

    We expect some number of workers to die, especially if we're using cheap
    "spot-priced" instances. Handle their death.
    """
    scrub_workers()



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Useful support code for Debusine views."""

import abc
import json
from dataclasses import dataclass
from typing import Any, override

import pygments
import pygments.formatters
import pygments.lexers
import yaml
from django.core.serializers.json import DjangoJSONEncoder
from django.template.defaultfilters import filesizeformat
from django.utils.html import format_html
from django.utils.safestring import SafeString

from debusine.db.models import FileInArtifact


class UIDumper(yaml.SafeDumper):
    """A YAML dumper that represents multi-line strings in the literal style."""

    @override
    def represent_scalar(
        self, tag: str, value: Any, style: str | None = None
    ) -> yaml.ScalarNode:
        """Represent multi-line strings in the literal style."""
        if style is None and "\n" in value:
            style = "|"
        return super().represent_scalar(tag, value, style=style)


def syntax_highlight(lexer_name: str, data: str) -> SafeString:
    """Return syntax-highlighted HTML for data using Pygments lexer."""
    formatter = pygments.formatters.HtmlFormatter(
        cssclass="formatted-content", linenos=False
    )
    lexer = pygments.lexers.get_lexer_by_name(lexer_name)
    return SafeString(pygments.highlight(data, lexer, formatter))


def highlight_as_yaml(data: Any, sort_keys: bool = True) -> SafeString:
    """Return syntax-highlighted HTML for data serialized as YAML."""
    dump = yaml.dump(data, Dumper=UIDumper, sort_keys=sort_keys)
    return syntax_highlight("yaml", dump)


def highlight_as_json(data: Any) -> SafeString:
    """Return syntax-highlighted HTML for data serialized as JSON."""
    dump = json.dumps(data, cls=DjangoJSONEncoder, indent=2)
    return syntax_highlight("json", dump)


@dataclass(slots=True)
class FileInArtifactError(Exception):
    """Base class for FileInArtifact errors."""

    file_in_artifact: FileInArtifact

    @abc.abstractmethod
    def to_html(self) -> SafeString:
        """Return a user-facing HTML error message."""


@dataclass(slots=True)
class IncompleteFileInArtifactError(FileInArtifactError):
    """Attempted to read an incomplete file."""

    @override
    def to_html(self) -> SafeString:
        """Return an HTML message indicating the file is still uploading."""
        return format_html(
            'File "{path}" has not completed uploading.',
            path=self.file_in_artifact.path,
        )


@dataclass(slots=True)
class TooBigFileInArtifactError(FileInArtifactError):
    """Attempted to read a too big file."""

    max_size: int

    @override
    def to_html(self) -> SafeString:
        """Return an HTML message with links to download or view the file."""
        return format_html(
            'File too big to display (size: {size} maximum size: {max_size}). '
            '<a href="{download_url}">Download</a> or '
            '<a href="{raw_url}">view</a> it raw.',
            download_url=self.file_in_artifact.get_absolute_url_download(),
            raw_url=self.file_in_artifact.get_absolute_url_raw(),
            size=filesizeformat(self.file_in_artifact.file.size),
            max_size=filesizeformat(self.max_size),
        )


def read_file(file_in_artifact: FileInArtifact, *, max_file_size: int) -> bytes:
    """
    Read contents if file is completed and size <= max_file_size.

    Raise exception if not possible to read the file.
    """
    if not file_in_artifact.complete:
        raise IncompleteFileInArtifactError(file_in_artifact=file_in_artifact)

    file = file_in_artifact.file
    if file.size > max_file_size:
        raise TooBigFileInArtifactError(
            file_in_artifact=file_in_artifact,
            max_size=max_file_size,
        )

    scope = file_in_artifact.artifact.workspace.scope

    with (
        scope.download_file_backend(file)
        .get_entry(file)
        .get_temporary_local_path() as path
    ):
        return path.read_bytes()



# Generated by Django 3.2.19 on 2024-05-30 12:23

import django.db.models.expressions
from django.db import migrations, models


class Migration(migrations.Migration):
    initial = True

    dependencies = []

    operations = [
        migrations.CreateModel(
            name='AuditLog',
            fields=[
                (
                    'id',
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                (
                    'purpose',
                    models.CharField(
                        choices=[('uefi', 'UEFI Secure Boot')], max_length=7
                    ),
                ),
                ('fingerprint', models.CharField(max_length=64)),
                (
                    'event',
                    models.CharField(
                        choices=[('generate', 'Generate'), ('sign', 'Sign')],
                        max_length=8,
                    ),
                ),
                ('data', models.JSONField()),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('created_by_work_request_id', models.IntegerField()),
            ],
        ),
        migrations.CreateModel(
            name='Key',
            fields=[
                (
                    'id',
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name='ID',
                    ),
                ),
                (
                    'purpose',
                    models.CharField(
                        choices=[('uefi', 'UEFI Secure Boot')], max_length=7
                    ),
                ),
                ('fingerprint', models.CharField(max_length=64)),
                ('private_key', models.JSONField()),
                ('public_key', models.BinaryField()),
                ('created_at', models.DateTimeField(auto_now_add=True)),
                ('updated_at', models.DateTimeField(blank=True, null=True)),
            ],
        ),
        migrations.AddConstraint(
            model_name='key',
            constraint=models.UniqueConstraint(
                fields=('purpose', 'fingerprint'),
                name='signing_key_unique_purpose_fingerprint',
            ),
        ),
        migrations.AddIndex(
            model_name='auditlog',
            index=models.Index(
                django.db.models.expressions.F('purpose'),
                django.db.models.expressions.F('fingerprint'),
                name='signing_auditlog_key_idx',
            ),
        ),
        migrations.AddIndex(
            model_name='auditlog',
            index=models.Index(
                django.db.models.expressions.F('created_by_work_request_id'),
                name='signing_auditlog_wr_idx',
            ),
        ),
    ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the management command delete_tokens."""

import io

from django.core.management import CommandError

from debusine.db.models import Token
from debusine.django.management.tests import call_command
from debusine.server.management.commands.tests.utils import TabularOutputTests
from debusine.test.django import TestCase


class DeleteTokensCommandTests(TabularOutputTests, TestCase):
    """Tests for the delete_tokens command."""

    def test_delete_tokens(self) -> None:
        """delete_tokens deletes the token and prints the deleted key."""
        token = self.playground.create_user_token()

        with self.assertPrintsTable() as output:
            call_command("delete_tokens", "--yes", "--token", token.key)
        self.assertEqual(output.col(0), [token.hash])

    def test_delete_tokens_no_tokens(self) -> None:
        """delete_tokens does not exist: raise CommandError."""
        with self.assertRaisesRegex(
            CommandError, "^There are no tokens to be deleted$"
        ) as exc:
            call_command(
                'delete_tokens',
                '--token',
                '9deefd915ecd1009bf7598c1d4acf9a3bbfb8bd9e0c08b4bdc9',
            )

        self.assertEqual(exc.exception.returncode, 3)

    def test_delete_tokens_no_tokens_force(self) -> None:
        """delete_tokens returns exit code 0 and prints error message."""
        stdout, _, _ = call_command(
            'delete_tokens',
            '--force',
            '--token',
            '9deefd915ecd1009bf7598c1d4acf9a3bbfb8bd9e0c08b4bdc93d099b9a38aa2',
        )

        self.assertEqual('There are no tokens to be deleted\n', stdout)

    def test_delete_tokens_confirmation(self) -> None:
        """delete_tokens doesn't delete the token (user does not confirm)."""
        token = self.playground.create_user_token()

        call_command(
            "delete_tokens", '--token', token.key, stdin=io.StringIO("N\n")
        )

        self.assertQuerySetEqual(Token.objects.filter(hash=token.hash), [token])

    def test_delete_tokens_confirmed(self) -> None:
        """delete_tokens delete the token (confirmed by the user)."""
        token = self.playground.create_user_token()

        self.assertEqual(Token.objects.filter(hash=token.hash).count(), 1)

        call_command(
            "delete_tokens", '--token', token.key, stdin=io.StringIO("y\n")
        )

        self.assertEqual(Token.objects.filter(hash=token.hash).count(), 0)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Config reader for the debusine workers."""

import logging
import os
from collections.abc import Sequence
from configparser import ConfigParser, NoOptionError, NoSectionError
from pathlib import Path
from typing import Literal, NoReturn, overload
from urllib.parse import urlparse

from debusine.utils import atomic_writer
from debusine.utils.urls import DEFAULT_REQUEST_TIMEOUT, parse_timeout

logger = logging.getLogger(__name__)


class ConfigHandler(ConfigParser):
    """Handle debusine worker configuration file (.ini format) and tokens."""

    default_directories = [
        str(Path.home() / '.config/debusine/worker'),
        '/etc/debusine/worker',
    ]

    def __init__(
        self,
        *,
        directories: Sequence[str] | None = None,
        require_https: bool = False,
    ) -> None:
        """
        Initialize variables and reads the configuration file.

        :param directories: if None the default directories are used
          (/etc/debusine/worker and ~/.config/debusine/worker)
        :param require_https: if True, require the configured api-url to use
          HTTPS
        """
        super().__init__()

        self._token: str | None = None
        self._api_url: str | None = None

        if directories is None:
            directories = self.default_directories

        self._configuration_directory = self._choose_directory(directories)
        self._require_https = require_https

        self.active_configuration_file = os.path.join(
            self._configuration_directory, 'config.ini'
        )

        self._read_file_or_fail(self.active_configuration_file)

        self._token_file = os.path.join(self._configuration_directory, 'token')
        self._activation_token_file = os.path.join(
            self._configuration_directory, 'activation-token'
        )

    def _read_file_or_fail(self, config_filename: str) -> None:
        """Read config_filename or fails."""
        files_read = self.read(config_filename)

        if len(files_read) != 1:
            self._fail(f'Cannot read {config_filename}')

    @staticmethod
    def _fail(message: str) -> NoReturn:
        """
        Print message and fail (exits) the execution of the Worker.

        Use only for non-recoverable errors.

        :param message: message to be logged in fatal
        :raises SystemExit(3): to finish the execution of the worker
        """
        logger.fatal(message)
        raise SystemExit(3)

    @classmethod
    def _choose_directory(cls, directories: Sequence[str]) -> str:
        for possible_directory in directories:
            if os.path.isdir(possible_directory):
                return possible_directory

        cls._fail(f'Configuration directory cannot be found in: {directories}')

    @property
    def token(self) -> str | None:
        """
        Return cached token or read it from the file.

        When accessing the file checks that the permissions are correct.

        :raises SystemExit: if the permissions are too open or the file is not
          accessible.
        """
        if self._token is not None:
            return self._token

        self._file_permissions_are_restricted_or_fail(self._token_file)

        try:
            with open(self._token_file) as f:
                self._token = f.read().rstrip('\n')
        except FileNotFoundError:
            pass
        except OSError as exc:
            self._fail(f'Cannot read {self._token_file}: {exc}')

        return self._token

    @property
    def activation_token(self) -> str | None:
        """
        Return activation token or read it from the file.

        When accessing the file checks that the permissions are correct.

        The activation token is deliberately uncached, since it is only used
        once during worker registration.

        :raises SystemExit: if the permissions are too open or the file is not
          accessible.
        """
        self._file_permissions_are_restricted_or_fail(
            self._activation_token_file
        )

        try:
            with open(self._activation_token_file) as f:
                return f.read().rstrip('\n')
        except FileNotFoundError:
            return None
        except OSError as exc:
            self._fail(f'Cannot read {self._activation_token_file}: {exc}')

    @property
    def api_url(self) -> str:
        """
        Return the value of api-url key in [General]. Caches it.

        :raises SystemExit: (via self._get_section_key) if
          [General] api-url is not in the configuration file, or (via
          self._fail) if this worker type requires HTTPS and the configured
          api-url does not use HTTPS
        """
        if self._api_url is not None:
            return self._api_url

        self._api_url = self._get_section_key('General', 'api-url')

        if self._require_https and urlparse(self._api_url).scheme != "https":
            self._fail(
                f'api-url in {self.active_configuration_file} does not use '
                f'HTTPS: {self._api_url}'
            )

        return self._api_url

    @property
    def log_level(self) -> str | None:
        """Return the log-level or None."""
        return self._get_section_key('General', 'log-level', required=False)

    @property
    def working_directory(self) -> Path:
        """Return the directory where the worker will execute jobs."""
        workdir = self._get_section_key(
            'General', 'working-directory', required=False
        )
        homedir = os.environ.get('HOME')
        if workdir:
            return Path(workdir)
        elif homedir:
            return Path(homedir) / 'workdir'
        else:
            return Path('/tmp')

    @property
    def environment(self) -> dict[str, str]:
        """Extra environment variables for tasks."""
        env = {}
        section = "Environment"
        if section in self:
            for key in self[section]:
                # Force the environment variable name to upper-case.
                # ConfigParser converts to lower-case by default, and changing
                # this would affect parsing other sections of the file
                env[key.upper()] = self.get(section, key)
        return env

    @property
    def http_timeout(self) -> int | tuple[int, int]:
        """Return the http-timeout from the config or the default."""
        raw_timeout = self._get_section_key(
            'General', 'http-timeout', required=False
        )
        if raw_timeout:
            try:
                return parse_timeout(raw_timeout)
            except ValueError:
                logger.warning(
                    "Invalid http-timeout value in worker configuration: %r. "
                    "Falling back to default (%s).",
                    raw_timeout,
                    DEFAULT_REQUEST_TIMEOUT,
                )
        return DEFAULT_REQUEST_TIMEOUT

    @overload
    def _get_section_key(
        self, section: str, key: str, *, required: Literal[True] = True
    ) -> str: ...

    @overload
    def _get_section_key(
        self, section: str, key: str, *, required: Literal[False]
    ) -> str | None: ...

    def _get_section_key(
        self, section: str, key: str, *, required: bool = True
    ) -> str | None:
        """
        Return value for the key in section.

        :param section: name of the section.
        :param key: key to retrieve.
        :param required: if it is True, if section or key does not exist
          it logs an error and exits.
        :raises SystemExit: via self._fail() if a required section or key
          is not found in the configuration file.
        """
        try:
            value = self.get(section, key)
        except NoSectionError:
            if required is False:
                return None
            self._fail(
                f'Missing required section "{section}" in the configuration '
                f'file ({self.active_configuration_file})'
            )
        except NoOptionError:
            if required is False:
                return None
            self._fail(
                f'Missing required key "{key}" in General section '
                f'in the configuration file '
                f'({self.active_configuration_file})'
            )

        return value

    def _file_permissions_are_restricted_or_fail(self, file_path: str) -> None:
        try:
            valid = not bool(os.stat(file_path).st_mode & 0o077)
        except FileNotFoundError:
            # File is not found: the file permissions are not too open
            return

        if not valid:
            self._fail(
                f'Permission too open for {file_path}. '
                f'Make sure that the file is not accessible by '
                f'group or others'
            )

    def write_token(self, token: str) -> None:
        """Write token into the token file."""
        try:
            with atomic_writer(self._token_file, mode="w", chmod=0o600) as fd:
                fd.write(token)
        except OSError as exc:
            self._fail(f'Cannot open token file: {exc}')

        self._token = token

    def validate_config_or_fail(self) -> None:
        """Validate the current configuration."""
        # The properties self.api_url, self.token, and self.activation_token
        # can call self._fail if some section or key does not exist, or if
        # the token file exists but is not accessible (permissions), etc.
        self.api_url
        self.token
        self.activation_token



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for scope handling in views."""

import base64
import datetime as dt
from collections.abc import Awaitable, Callable
from typing import Any, Protocol, cast
from unittest import mock

import django.http
from django.conf import settings
from django.contrib.auth.models import AnonymousUser
from django.core.exceptions import ImproperlyConfigured
from django.test import RequestFactory
from django.utils import timezone
from django.utils.functional import SimpleLazyObject
from rest_framework import HTTP_HEADER_ENCODING

from debusine.db.context import ContextConsistencyError, context
from debusine.db.models import Token, User, WorkRequest
from debusine.db.permissioncontext import PermissionUser
from debusine.db.playground import scenarios
from debusine.server.middlewares.scopes import (
    AuthorizationMiddleware,
    ScopeMiddleware,
    TOKEN_LAST_SEEN_AT_UPDATE_INTERVAL,
)
from debusine.test.django import TestCase

#: Singleton response used to check if a middleware called get_response()
MOCK_RESPONSE = django.http.HttpResponse()


class MiddlewareProtocol(Protocol):
    """A protocol defining our expectations of middlewares."""

    def __init__(
        self,
        get_response: Callable[
            [django.http.HttpRequest], django.http.HttpResponse
        ],
    ) -> None:
        """Instantiate the middleware."""

    def __call__(
        self, request: django.http.HttpRequest
    ) -> django.http.HttpResponseBase | Awaitable[django.http.HttpResponseBase]:
        """Call the middleware."""


class MiddlewareTestMixin[M: MiddlewareProtocol](TestCase):
    """Common functions to test middlewares."""

    middleware_class: type[M]

    scenario = scenarios.DefaultScopeUser()

    def get_middleware(self) -> M:
        """Instantiate a test middleware."""
        return self.middleware_class(
            cast(
                Callable[[django.http.HttpRequest], django.http.HttpResponse],
                lambda _: MOCK_RESPONSE,
            )
        )

    def request_for_path(
        self, path_info: str, **kwargs: Any
    ) -> django.http.HttpRequest:
        """Configure a request for path_info with the middleware."""
        request = RequestFactory().get(path_info, **kwargs)
        self.get_middleware()(request)
        return request


class ScopeMiddlewareTests(MiddlewareTestMixin[ScopeMiddleware]):
    """Test ScopeMiddleware."""

    middleware_class = ScopeMiddleware

    def test_get_scope(self) -> None:
        """Test get_scope."""
        other_scope = self.playground.get_or_create_scope("other")
        mw = self.get_middleware()
        self.assertEqual(
            mw.get_scope(self.scenario.scope.name), self.scenario.scope
        )
        self.assertEqual(mw.get_scope("other"), other_scope)
        with self.assertRaises(django.http.Http404):
            mw.get_scope("nonexistent")

    def test_request_setup_debusine(self) -> None:
        """Test request setup with the fallback scope."""
        request = self.request_for_path(f"/{settings.DEBUSINE_DEFAULT_SCOPE}/")
        self.assertEqual(getattr(request, "urlconf"), "debusine.project.urls")
        self.assertEqual(
            getattr(request, "debusine_scope"), self.scenario.scope
        )

    def test_request_setup_api(self) -> None:
        """Test request setup for API calls."""
        request = self.request_for_path("/api")
        self.assertEqual(getattr(request, "urlconf"), "debusine.project.urls")
        self.assertEqual(
            getattr(request, "debusine_scope"), self.scenario.scope
        )

    def test_request_setup_api_scope_in_header(self) -> None:
        """Test request setup for API calls with scope in header."""
        other_scope = self.playground.get_or_create_scope("other")
        request = self.request_for_path(
            "/api", headers={"X-Debusine-Scope": "other"}
        )
        self.assertEqual(
            getattr(request, "urlconf"), "debusine.server._urlconfs.other"
        )
        self.assertEqual(getattr(request, "debusine_scope"), other_scope)

    def test_request_setup_scope(self) -> None:
        """Test request setup with a valid scope."""
        other_scope = self.playground.get_or_create_scope("other")
        request = self.request_for_path("/other/")
        self.assertEqual(
            getattr(request, "urlconf"),
            "debusine.server._urlconfs.other",
        )
        self.assertEqual(getattr(request, "debusine_scope"), other_scope)

    def test_request_setup_wrong_scope(self) -> None:
        """Test request setup with an invalid scope."""
        with self.assertRaises(django.http.Http404):
            self.request_for_path("/wrongscope/")


class AuthorizationMiddlewareTests(
    MiddlewareTestMixin[AuthorizationMiddleware]
):
    """Test AuthorizationMiddleware."""

    middleware_class = AuthorizationMiddleware

    def make_request(
        self,
        user: User | None = None,
        token: Token | str | None = None,
        *,
        secure: bool = True,
    ) -> django.http.HttpRequest:
        """Create a request."""
        headers = {}
        match token:
            case Token():
                headers["token"] = token.key
            case str():
                headers["token"] = token
        request = RequestFactory().get("/", headers=headers, secure=secure)
        request.user = user or AnonymousUser()
        request.debusine_scope = self.scenario.scope  # type: ignore[attr-defined]
        return request

    def make_request_archive_view(
        self,
        user: User | None = None,
        token: Token | str | None = None,
        *,
        secure: bool = True,
    ) -> django.http.HttpRequest:
        """Create a request for archive views."""
        request = self.make_request(user, token, secure=secure)
        setattr(request, "is_archive_view", True)
        return request

    def add_basic_auth_header(
        self, request: django.http.HttpRequest, username: str, password: str
    ) -> None:
        """Add an http basic auth header to the request."""
        credentials = f"{username}:{password}"
        base64_credentials = base64.b64encode(
            credentials.encode(HTTP_HEADER_ENCODING)
        ).decode("ascii")
        request.META["HTTP_AUTHORIZATION"] = f"Basic {base64_credentials}"

    def add_basic_auth_token(
        self, request: django.http.HttpRequest, token: Token
    ) -> None:
        """Pass the user token as a basic auth header."""
        assert token.user is not None
        self.add_basic_auth_header(request, token.user.username, token.key)

    def call_middleware(
        self, request: django.http.HttpRequest
    ) -> django.http.HttpResponse:
        """Process the request through the middleware."""
        mw = self.get_middleware()
        response = mw(request)
        assert isinstance(response, django.http.HttpResponse)
        return response

    def assertSetsUserNone(
        self, request: django.http.HttpRequest
    ) -> django.http.HttpResponse:
        """
        Ensure that the middleware sets user context to None.

        :returns: the response from the middleware.
        """
        response = self.call_middleware(request)
        self.assertPermissionContextNotSet()
        return response

    def assertSetsUser(
        self, request: django.http.HttpRequest, user: PermissionUser
    ) -> django.http.HttpResponse:
        """
        Ensure that the middleware sets user context to the given user.

        :returns: the response from the middleware.
        """
        response = self.call_middleware(request)
        self.assertEqual(context.user, user)
        return response

    def assertResponse403(
        self, response: django.http.HttpResponse, message: str
    ) -> None:
        """Ensure the response is a 403 with the given message."""
        assert isinstance(response, django.http.HttpResponse)
        self.assertEqual(response.status_code, 403)
        self.assertEqual(response.content.decode(), message)

    def assertPermissionContextNotSet(self) -> None:
        """Ensure the permission context has not been set."""
        self.assertIsNone(context._permission_context.get())

    def test_requires_scope_in_request(self) -> None:
        request = RequestFactory().get("/")
        request.user = AnonymousUser()
        with self.assertRaisesRegex(
            ImproperlyConfigured,
            "AuthorizationMiddleware needs to be called after ScopeMiddleware",
        ):
            self.call_middleware(request)

    def test_allowed(self) -> None:
        """Test the allowed case."""
        # There currently is no permission predicate implemented for scope
        # visibility, only a comment placeholder in Context.set_user.
        request = self.make_request(user=self.scenario.user)
        response = self.assertSetsUser(request, self.scenario.user)
        self.assertIs(response, MOCK_RESPONSE)

    def test_forbidden(self) -> None:
        """Test the forbidden case."""
        # There currently is no permission predicate implemented for scope
        # visibility, only a comment placeholder in Context.set_user.
        # Mock Context.set_user instead of a permission, to test handling of
        # failure to set it
        request = self.make_request(user=self.scenario.user)
        with mock.patch(
            "debusine.db.context.Context.set_permission_context_scoped",
            side_effect=ContextConsistencyError("expected fail"),
        ):
            response = self.assertSetsUserNone(request)
        self.assertResponse403(response, "expected fail")

    def test_evaluates_lazy_object(self) -> None:
        """Lazy objects are evaluated before storing them in the context."""
        mw = self.get_middleware()
        request = RequestFactory().get("/")
        request.user = SimpleLazyObject(lambda: self.scenario.user)  # type: ignore[assignment]
        request.debusine_scope = self.scenario.scope  # type: ignore[attr-defined]
        self.assertIs(mw(request), MOCK_RESPONSE)
        with mock.patch.object(User.objects, "get", side_effect=RuntimeError):
            self.assertEqual(context.user, self.scenario.user)

    def test_user_token(self) -> None:
        """Check handling of user tokens."""
        token = self.scenario.create_user_token()
        request = self.make_request(token=token)
        response = self.assertSetsUser(request, self.scenario.user)
        self.assertIs(response, MOCK_RESPONSE)
        self.assertEqual(getattr(request, "_debusine_token"), token)
        self.assertEqual(context.user_token, token)
        self.assertIsNone(context.external_user_token)
        self.assertIsNone(context.pc.worker_token)
        self.assertIsNone(context.pc.worker_activation_token)
        self.assertIsNone(context.pc.work_request_token)

    def test_user_token_inactive_user(self) -> None:
        """Inactive users are not allowed."""
        token = self.scenario.create_user_token()
        assert token.user is not None
        token.user.is_active = False
        token.user.save()
        request = self.make_request(token=token)
        response = self.assertSetsUserNone(request)
        assert isinstance(response, django.http.HttpResponse)
        self.assertResponse403(response, "user token has an inactive user")

    def test_user_token_logged_in(self) -> None:
        """Check handling of user tokens while logged in."""
        token = self.scenario.create_user_token()
        request = self.make_request(token=token, user=token.user)
        response = self.assertSetsUserNone(request)
        self.assertResponse403(
            response, "cannot use both Django and user token authentication"
        )
        self.assertPermissionContextNotSet()

    def test_external_user_token(self) -> None:
        """Check handling of external user tokens."""
        token = self.playground.create_external_user_token()
        request = self.make_request(token=token)
        response = self.assertSetsUser(request, AnonymousUser())
        self.assertIs(response, MOCK_RESPONSE)
        self.assertEqual(getattr(request, "_debusine_token"), token)
        self.assertEqual(context.external_user_token, token)
        self.assertIsNone(context.pc.work_request_token)
        self.assertIsNone(context.pc.worker_activation_token)
        self.assertIsNone(context.pc.worker_token)
        self.assertIsNone(context.user_token)

    def test_external_user_token_logged_in(self) -> None:
        """Check handling of external user tokens while logged in."""
        token = self.playground.create_external_user_token()
        request = self.make_request(
            token=token, user=self.playground.get_default_user()
        )
        response = self.assertSetsUserNone(request)
        self.assertResponse403(
            response, "cannot use both Django and user token authentication"
        )
        self.assertPermissionContextNotSet()

    def test_user_token_forbidden(self) -> None:
        """Test the forbidden case for user token."""
        token = self.scenario.create_user_token()
        request = self.make_request(token=token)

        # There currently is no permission predicate implemented for scope
        # visibility, only a comment placeholder in Context.set_user.
        # Mock Context.set_user instead of a permission, to test handling of
        # failure to set it
        with mock.patch(
            "debusine.db.context.Context.set_permission_context_scoped",
            side_effect=ContextConsistencyError("expected fail"),
        ):
            response = self.assertSetsUserNone(request)
        self.assertResponse403(response, "expected fail")

    def test_worker_token(self) -> None:
        """Check handling of worker tokens."""
        token = self.playground.create_worker_token()
        request = self.make_request(token=token)
        response = self.call_middleware(request)
        self.assertEqual(context.user, AnonymousUser())
        self.assertIs(response, MOCK_RESPONSE)
        self.assertEqual(getattr(request, "_debusine_token"), token)
        self.assertEqual(context.pc.worker_token, token)
        self.assertIsNone(context.external_user_token)
        self.assertIsNone(context.pc.work_request_token)
        self.assertIsNone(context.pc.worker_activation_token)
        self.assertIsNone(context.user_token)

    def test_worker_token_disabled(self) -> None:
        """Check handling of disabled tokens."""
        token = self.playground.create_worker_token()
        token.disable()
        request = self.make_request(token=token)
        response = self.assertSetsUser(request, AnonymousUser())
        self.assertIs(response, MOCK_RESPONSE)
        self.assertIsNone(getattr(request, "_debusine_token"))
        self.assertIsNone(context.user_token)
        self.assertIsNone(context.external_user_token)
        self.assertIsNone(context.pc.worker_activation_token)
        self.assertIsNone(context.pc.worker_token)
        self.assertIsNone(context.pc.work_request_token)

    def test_worker_activation_token(self) -> None:
        """Check handling of worker activation tokens."""
        token = self.playground.create_worker_activation_token()
        request = self.make_request(token=token)
        response = self.assertSetsUser(request, AnonymousUser())
        self.assertIs(response, MOCK_RESPONSE)
        self.assertEqual(getattr(request, "_debusine_token"), token)
        self.assertEqual(context.pc.worker_activation_token, token)
        self.assertIsNone(context.pc.worker_token)
        self.assertIsNone(context.pc.work_request_token)

    def test_worker_activation_token_disabled(self) -> None:
        """Check handling of disabled worker activation tokens."""
        token = self.playground.create_worker_activation_token(enabled=False)
        request = self.make_request(token=token)
        response = self.assertSetsUser(request, AnonymousUser())
        self.assertIs(response, MOCK_RESPONSE)
        self.assertIsNone(getattr(request, "_debusine_token"))
        self.assertIsNone(context.user_token)
        self.assertIsNone(context.external_user_token)
        self.assertIsNone(context.pc.worker_activation_token)
        self.assertIsNone(context.pc.worker_token)
        self.assertIsNone(context.pc.work_request_token)

    def test_work_request_token(self) -> None:
        """Check handling of work request worker tokens."""
        token = self.playground.create_work_request_token()
        assert token.work_request is not None
        request = self.make_request(token=token)
        response = self.call_middleware(request)
        self.assertEqual(context.user, token.work_request.created_by)
        self.assertIs(response, MOCK_RESPONSE)
        self.assertEqual(getattr(request, "_debusine_token"), token)
        self.assertEqual(context.pc.work_request_token, token)
        self.assertIsNone(context.external_user_token)
        self.assertIsNone(context.pc.worker_activation_token)
        self.assertIsNone(context.pc.worker_token)
        self.assertIsNone(context.user_token)

    def test_work_request_worker_token_status(self) -> None:
        """Work Request tokens must belong to running work requests."""
        work_request = self.playground.create_worker_task(
            mark_running=True, assign_new_worker=True
        )
        token = self.playground.create_work_request_token(work_request)

        for status, expected in (
            (WorkRequest.Statuses.PENDING, False),
            (WorkRequest.Statuses.RUNNING, True),
            (WorkRequest.Statuses.COMPLETED, False),
            (WorkRequest.Statuses.ABORTED, False),
        ):
            with self.subTest(status=str(status)), context.local():
                # We don't use advance_work_request, as we do not want the
                # token to be deleted, once the task is done.
                work_request.status = status
                work_request.save()

                request = self.make_request(token=token)
                response = self.call_middleware(request)
                if not expected:
                    self.assertPermissionContextNotSet()
                    self.assertResponse403(
                        response, "work request is not running"
                    )
                else:
                    self.assertEqual(context.user, work_request.created_by)

    def test_token_invalid(self) -> None:
        """Check handling of invalid tokens."""
        mw = self.get_middleware()
        request = self.make_request(token="invalid")
        response = mw(request)
        self.assertIs(response, MOCK_RESPONSE)
        self.assertIsNone(getattr(request, "_debusine_token"))
        self.assertIsNone(context.user_token)
        self.assertIsNone(context.external_user_token)
        self.assertIsNone(context.pc.worker_activation_token)
        self.assertIsNone(context.pc.worker_token)
        self.assertIsNone(context.pc.work_request_token)

    def test_count_queries(self) -> None:
        """Check the number of queries done to authenticate."""
        token = self.playground.create_worker_token()
        request = self.make_request(token=token)
        with self.assertNumQueries(2):
            response = self.call_middleware(request)
            self.assertEqual(context.user, AnonymousUser())
        self.assertIs(response, MOCK_RESPONSE)

    def test_update_last_seen_at(self) -> None:
        """Request with a token updates last_seen_at."""
        before = timezone.now()
        token = self.scenario.create_user_token()
        self.assertIsNone(token.last_seen_at)
        request = self.make_request(token=token)
        response = self.assertSetsUser(request, self.scenario.user)
        self.assertIs(response, MOCK_RESPONSE)

        # token.last_seen_at was updated
        token.refresh_from_db()
        assert token.last_seen_at is not None
        last_seen_at = token.last_seen_at
        self.assertGreaterEqual(token.last_seen_at, before)
        self.assertLessEqual(token.last_seen_at, timezone.now())

        # token.last_seen_at is not updated (it needs at least
        # TOKEN_LAST_SEEN_AT_UPDATE_INTERVAL)
        request = self.make_request(token=token)
        self.assertSetsUser(request, self.scenario.user)
        token.refresh_from_db()
        self.assertEqual(token.last_seen_at, last_seen_at)

        # If the token.last_seen_at is older than
        # TOKEN_LAST_SEEN_AT_UPDATE_INTERVAL: it gets updated
        stale_last_seen_at = (
            timezone.now()
            - TOKEN_LAST_SEEN_AT_UPDATE_INTERVAL
            - dt.timedelta(seconds=1)
        )
        token.last_seen_at = stale_last_seen_at
        token.save()

        before_update = timezone.now()
        request = self.make_request(token=token)
        self.assertSetsUser(request, self.scenario.user)

        token.refresh_from_db()
        self.assertGreaterEqual(token.last_seen_at, before_update)
        self.assertLessEqual(token.last_seen_at, timezone.now())

    def test_archive_views_http_no_basic_auth(self) -> None:
        request = self.make_request_archive_view(secure=False)
        response = self.assertSetsUser(request, AnonymousUser())
        self.assertIs(response, MOCK_RESPONSE)
        self.assertIsNone(context.user_token)
        self.assertIsNone(context.external_user_token)
        self.assertIsNone(context.pc.worker_activation_token)
        self.assertIsNone(context.pc.worker_token)
        self.assertIsNone(context.pc.work_request_token)

    def test_token_basic_auth_not_debian_archive_view(self) -> None:
        token = self.scenario.create_user_token()
        request = self.make_request()
        self.add_basic_auth_token(request, token)
        response = self.assertSetsUser(request, AnonymousUser())
        self.assertIs(response, MOCK_RESPONSE)
        self.assertIsNone(context.user_token)
        self.assertIsNone(context.external_user_token)
        self.assertIsNone(context.pc.worker_activation_token)
        self.assertIsNone(context.pc.worker_token)
        self.assertIsNone(context.pc.work_request_token)

    def test_token_basic_auth_authorization_not_basic(self) -> None:
        request = self.make_request_archive_view()
        request.META["HTTP_AUTHORIZATION"] = "Token foo"
        response = self.assertSetsUser(request, AnonymousUser())
        self.assertIs(response, MOCK_RESPONSE)
        self.assertIsNone(context.user_token)
        self.assertIsNone(context.external_user_token)
        self.assertIsNone(context.pc.worker_activation_token)
        self.assertIsNone(context.pc.worker_token)
        self.assertIsNone(context.pc.work_request_token)

    def test_token_basic_auth_no_basic_credentials(self) -> None:
        request = self.make_request_archive_view()
        request.META["HTTP_AUTHORIZATION"] = "Basic"
        response = self.assertSetsUserNone(request)
        self.assertResponse403(response, "No credentials provided")
        self.assertPermissionContextNotSet()

    def test_token_basic_auth_credentials_malformed(self) -> None:
        for credentials in (
            "foo bar",
            "foo",
            # No colon in credentials
            base64.b64encode("foobar".encode(HTTP_HEADER_ENCODING)).decode(
                "ascii"
            ),
            # Invalid credentials, encoded in invalid UTF-8
            base64.b64encode(b"\x80").decode("ascii"),
        ):
            with self.subTest(credentials=repr(credentials)):
                request = self.make_request_archive_view()
                request.META["HTTP_AUTHORIZATION"] = f"Basic {credentials}"
                response = self.assertSetsUserNone(request)
                self.assertResponse403(
                    response, "Credentials string is malformed"
                )
                self.assertPermissionContextNotSet()

    def test_token_basic_auth_not_https(self) -> None:
        token = self.scenario.create_user_token()
        request = self.make_request_archive_view(secure=False)
        self.add_basic_auth_token(request, token)
        response = self.assertSetsUserNone(request)
        self.assertResponse403(
            response,
            "Authenticated access is only allowed using HTTPS",
        )
        self.assertPermissionContextNotSet()

    def test_token_basic_auth_nonexistent_token(self) -> None:
        request = self.make_request_archive_view()
        self.add_basic_auth_header(
            request, self.scenario.user.username, "invalid-token"
        )
        response = self.assertSetsUserNone(request)
        self.assertResponse403(response, "Invalid token")
        self.assertPermissionContextNotSet()

    def test_token_basic_auth_worker_token(self) -> None:
        token = self.playground.create_worker_token()
        request = self.make_request_archive_view()
        self.add_basic_auth_header(
            request, self.scenario.user.username, token.key
        )
        response = self.assertSetsUserNone(request)
        self.assertResponse403(response, "Invalid token")
        self.assertPermissionContextNotSet()

    def test_token_basic_auth_worker_activation_token(self) -> None:
        token = self.playground.create_worker_activation_token()
        request = self.make_request_archive_view()
        self.add_basic_auth_header(
            request, self.scenario.user.username, token.key
        )
        response = self.assertSetsUserNone(request)
        self.assertResponse403(response, "Invalid token")
        self.assertPermissionContextNotSet()

    def test_token_basic_auth_wrong_user(self) -> None:
        user = self.playground.create_user("test")
        token = self.playground.create_user_token(user=user)
        request = self.make_request_archive_view()
        self.add_basic_auth_header(
            request, self.scenario.user.username, token.key
        )
        response = self.assertSetsUserNone(request)
        self.assertResponse403(response, "Invalid token")
        self.assertPermissionContextNotSet()

    def test_token_basic_auth_expired_user_token(self) -> None:
        token = self.playground.create_user_token(
            expire_at=timezone.now() - dt.timedelta(seconds=1)
        )
        request = self.make_request_archive_view()
        self.add_basic_auth_token(request, token)
        response = self.assertSetsUserNone(request)
        self.assertResponse403(response, "Invalid token")
        self.assertPermissionContextNotSet()

    def test_token_basic_auth_valid_user_token(self) -> None:
        token = self.scenario.create_user_token()
        request = self.make_request_archive_view()
        self.add_basic_auth_token(request, token)
        response = self.assertSetsUser(request, self.scenario.user)
        self.assertIs(response, MOCK_RESPONSE)
        self.assertEqual(context.user_token, token)
        self.assertIsNone(context.external_user_token)
        self.assertIsNone(context.pc.worker_activation_token)
        self.assertIsNone(context.pc.worker_token)
        self.assertIsNone(context.pc.work_request_token)
